from datetime import datetime
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple
from typing import Union
from urllib.parse import urljoin

from requests.models import Response

from tecton._internals import errors
from tecton_core import conf
from tecton_core import http


class IngestionClient:
    def __init__(self):
        self.ingestion_url: str = self.get_ingestion_url()

    def ingest(
        self,
        workspace_name: str,
        push_source_name: str,
        ingestion_records: Sequence[Dict[str, Any]],
        dry_run: bool = False,
    ) -> Tuple[int, str, Union[Dict[str, Any], None]]:
        http_request = {
            "workspace_name": workspace_name,
            "dry_run": dry_run,
            "records": {
                push_source_name: [
                    {
                        "record": self._serialize_record(ingestion_record),
                    }
                    for ingestion_record in ingestion_records
                ]
            },
        }
        response = http.session().post(self.ingestion_url, json=http_request, headers=self._prepare_headers())
        return self._extract_from_response(response)

    @staticmethod
    def get_ingestion_url() -> str:
        # This configuration variable is ultimately coming from MDS.
        # Broadly speaking, the `get_or_raise` method below will call a `_get` method in conf.py,
        # which will look through the hierarchy of sources for a configuration variable.
        # If INGESTION_SERVICE is found via one of the acceptable ones, then it's used.
        # These sources include MDS, which is populated via the `GetConfigs` RPC. These RPC only provides
        # configurations specified in the server-side code, and this includes INGESTION_SERVICE.
        # See tecton/_internals/conf.py and tecton/_internals/metadata_service.py for more details
        domain = conf.get_or_raise("INGESTION_SERVICE")
        # urljoin will *replace* the end of the base URL path with "/ingest"
        # unless the base path's URL path ends with a forward slash and
        # the url does not begin with a forward slash.
        # The domain name below has a forward slash appended as a protection
        # for cases where INGESTION_SERVICE does not include the trailing
        # forward slash.
        return urljoin(domain + "/", "ingest")

    @staticmethod
    def _extract_from_response(response: Response) -> Tuple[int, str, Union[Dict[str, Any], None]]:
        # NOTE: Responses not coming from our system will not necessarily be in JSON format

        if response.status_code != 200:
            error_message = f"Failed to ingest records: {response.reason}"
            try:
                response_data = response.json()
            except ValueError:
                response_data = None
                error_message += f"\nAdditionally, response content is not valid JSON: {response.text}"
            return response.status_code, error_message, response_data

        try:
            response_data = response.json()
        except ValueError:
            error_message = f"Response content is not valid JSON: {response.text}"
            return response.status_code, error_message, None
        return response.status_code, response.reason, response_data

    @staticmethod
    def _prepare_headers() -> Dict[str, str]:
        token = conf.get_or_none("TECTON_API_KEY")
        if not token:
            raise errors.FS_API_KEY_MISSING

        return {
            "authorization": f"Tecton-key {token}",
            "Content-Type": "application/json",
        }

    @staticmethod
    def _serialize_record(record: Dict[str, Any]) -> Dict[str, Any]:
        serialized_record = {}
        for k, v in record.items():
            if isinstance(v, datetime):
                serialized_record[k] = v.isoformat()
            else:
                serialized_record[k] = v
        return serialized_record



from tecton_core import conf
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_utils import get_first_input_node_of_class
from tecton_core.query.nodes import AsofJoinFullAggNode
from tecton_core.specs import TimeWindowSeriesSpec
from tecton_core.specs import create_time_window_spec_from_data_proto


def should_use_optimized_full_aggregate_node(qt_root: NodeRef) -> bool:
    """
    Decides whether to use optimized Full Aggregation algorithm (group-by based query) vs standard (window-based query).
    """
    if conf.get_or_none("DUCKDB_ENABLE_OPTIMIZED_FULL_AGG") is not None:
        # user enforced
        return conf.get_bool("DUCKDB_ENABLE_OPTIMIZED_FULL_AGG")

    full_agg_node = get_first_input_node_of_class(qt_root, AsofJoinFullAggNode)
    if not full_agg_node:
        return False

    if any(
        isinstance(create_time_window_spec_from_data_proto(feature.time_window), TimeWindowSeriesSpec)
        for feature in full_agg_node.fdw.fv_spec.aggregate_features
    ):
        # TimeWindowSeriesSpec is currently not supported by the optimized version
        return False

    return True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/calculation_node__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import data_type__client_pb2 as tecton__proto_dot_common_dot_data__type____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/common/calculation_node__client.proto\x12\x13tecton_proto.common\x1a+tecton_proto/common/data_type__client.proto\"\x12\n\x10NullLiteralValue\"\x96\x02\n\x0cLiteralValue\x12%\n\rfloat32_value\x18\x01 \x01(\x02H\x00R\x0c\x66loat32Value\x12%\n\rfloat64_value\x18\x02 \x01(\x01H\x00R\x0c\x66loat64Value\x12!\n\x0bint64_value\x18\x03 \x01(\x03H\x00R\nint64Value\x12\x1f\n\nbool_value\x18\x04 \x01(\x08H\x00R\tboolValue\x12#\n\x0cstring_value\x18\x05 \x01(\tH\x00R\x0bstringValue\x12\x46\n\nnull_value\x18\x06 \x01(\x0b\x32%.tecton_proto.common.NullLiteralValueH\x00R\tnullValueB\x07\n\x05value\"\x96\x01\n\tOperation\x12@\n\toperation\x18\x01 \x01(\x0e\x32\".tecton_proto.common.OperationTypeR\toperation\x12G\n\x08operands\x18\x02 \x03(\x0b\x32+.tecton_proto.common.AbstractSyntaxTreeNodeR\x08operands\"\x9c\x01\n\nWhenClause\x12I\n\tcondition\x18\x01 \x01(\x0b\x32+.tecton_proto.common.AbstractSyntaxTreeNodeR\tcondition\x12\x43\n\x06result\x18\x02 \x01(\x0b\x32+.tecton_proto.common.AbstractSyntaxTreeNodeR\x06result\"Q\n\nElseClause\x12\x43\n\x06result\x18\x01 \x01(\x0b\x32+.tecton_proto.common.AbstractSyntaxTreeNodeR\x06result\"\x95\x01\n\rCaseStatement\x12\x42\n\x0cwhen_clauses\x18\x01 \x03(\x0b\x32\x1f.tecton_proto.common.WhenClauseR\x0bwhenClauses\x12@\n\x0b\x65lse_clause\x18\x02 \x01(\x0b\x32\x1f.tecton_proto.common.ElseClauseR\nelseClause\"\x98\x03\n\x16\x41\x62stractSyntaxTreeNode\x12\x33\n\x05\x64type\x18\x01 \x01(\x0b\x32\x1d.tecton_proto.common.DataTypeR\x05\x64type\x12H\n\rliteral_value\x18\x02 \x01(\x0b\x32!.tecton_proto.common.LiteralValueH\x00R\x0cliteralValue\x12+\n\x10\x63olumn_reference\x18\x03 \x01(\tH\x00R\x0f\x63olumnReference\x12>\n\toperation\x18\x04 \x01(\x0b\x32\x1e.tecton_proto.common.OperationH\x00R\toperation\x12<\n\tdate_part\x18\x05 \x01(\x0e\x32\x1d.tecton_proto.common.DatePartH\x00R\x08\x64\x61tePart\x12K\n\x0e\x63\x61se_statement\x18\x06 \x01(\x0b\x32\".tecton_proto.common.CaseStatementH\x00R\rcaseStatementB\x07\n\x05value*\x8e\x02\n\rOperationType\x12\x19\n\x15OPERATION_UNSPECIFIED\x10\x00\x12\x0c\n\x08\x41\x44\x44ITION\x10\x01\x12\x0f\n\x0bSUBTRACTION\x10\x02\x12\x12\n\x0eMULTIPLICATION\x10\x03\x12\x0c\n\x08\x44IVISION\x10\x04\x12\x0c\n\x08\x43OALESCE\x10\x05\x12\r\n\tDATE_DIFF\x10\x06\x12\n\n\x06\x45QUALS\x10\x07\x12\x0e\n\nNOT_EQUALS\x10\x08\x12\r\n\tLESS_THAN\x10\t\x12\x14\n\x10LESS_THAN_EQUALS\x10\n\x12\x10\n\x0cGREATER_THAN\x10\x0b\x12\x17\n\x13GREATER_THAN_EQUALS\x10\x0c\x12\x07\n\x03\x41ND\x10\r\x12\x06\n\x02OR\x10\x0e\x12\x07\n\x03NOT\x10\x0f*\xc9\x01\n\x08\x44\x61tePart\x12\x19\n\x15\x44\x41TE_PART_UNSPECIFIED\x10\x00\x12\x07\n\x03\x44\x41Y\x10\x01\x12\t\n\x05MONTH\x10\x02\x12\x08\n\x04WEEK\x10\x03\x12\x08\n\x04YEAR\x10\x04\x12\n\n\x06SECOND\x10\x05\x12\x08\n\x04HOUR\x10\x06\x12\n\n\x06MINUTE\x10\x07\x12\x0e\n\nMILLENNIUM\x10\x08\x12\x0b\n\x07\x43\x45NTURY\x10\t\x12\n\n\x06\x44\x45\x43\x41\x44\x45\x10\n\x12\x0b\n\x07QUARTER\x10\x0b\x12\x10\n\x0cMILLISECONDS\x10\x0c\x12\x10\n\x0cMICROSECONDS\x10\rB?\n\x11\x63om.tecton.commonP\x01Z(github.com/tecton-ai/tecton_proto/common')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.calculation_node__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001Z(github.com/tecton-ai/tecton_proto/common'
  _OPERATIONTYPE._serialized_start=1380
  _OPERATIONTYPE._serialized_end=1650
  _DATEPART._serialized_start=1653
  _DATEPART._serialized_end=1854
  _NULLLITERALVALUE._serialized_start=120
  _NULLLITERALVALUE._serialized_end=138
  _LITERALVALUE._serialized_start=141
  _LITERALVALUE._serialized_end=419
  _OPERATION._serialized_start=422
  _OPERATION._serialized_end=572
  _WHENCLAUSE._serialized_start=575
  _WHENCLAUSE._serialized_end=731
  _ELSECLAUSE._serialized_start=733
  _ELSECLAUSE._serialized_end=814
  _CASESTATEMENT._serialized_start=817
  _CASESTATEMENT._serialized_end=966
  _ABSTRACTSYNTAXTREENODE._serialized_start=969
  _ABSTRACTSYNTAXTREENODE._serialized_end=1377
# @@protoc_insertion_point(module_scope)



from tecton_proto.databricks_api import jobs__client_pb2 as _jobs__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
FAILED: LibraryStatus
INSTALLED: LibraryStatus
INSTALLING: LibraryStatus
RESOLVING: LibraryStatus
RESTORED: LibraryStatus
SKIPPED: LibraryStatus
UNINSTALL_ON_RESTART: LibraryStatus

class ClusterStatusResponse(_message.Message):
    __slots__ = ["library_statuses"]
    LIBRARY_STATUSES_FIELD_NUMBER: _ClassVar[int]
    library_statuses: _containers.RepeatedCompositeFieldContainer[LibraryStatusEntry]
    def __init__(self, library_statuses: _Optional[_Iterable[_Union[LibraryStatusEntry, _Mapping]]] = ...) -> None: ...

class LibraryInstallRequest(_message.Message):
    __slots__ = ["cluster_id", "libraries"]
    CLUSTER_ID_FIELD_NUMBER: _ClassVar[int]
    LIBRARIES_FIELD_NUMBER: _ClassVar[int]
    cluster_id: str
    libraries: _containers.RepeatedCompositeFieldContainer[_jobs__client_pb2.RemoteLibrary]
    def __init__(self, cluster_id: _Optional[str] = ..., libraries: _Optional[_Iterable[_Union[_jobs__client_pb2.RemoteLibrary, _Mapping]]] = ...) -> None: ...

class LibraryStatusEntry(_message.Message):
    __slots__ = ["library", "status"]
    LIBRARY_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    library: _jobs__client_pb2.RemoteLibrary
    status: LibraryStatus
    def __init__(self, library: _Optional[_Union[_jobs__client_pb2.RemoteLibrary, _Mapping]] = ..., status: _Optional[_Union[LibraryStatus, str]] = ...) -> None: ...

class LibraryStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



import contextlib
import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import attrs
import pandas
import pyarrow
import pyarrow.dataset

from tecton_core import conf
from tecton_core import specs
from tecton_core.compute_mode import ComputeMode
from tecton_core.duckdb_factory import DuckDBConfig
from tecton_core.embeddings.model_artifacts import DEFAULT_MODEL_PROVIDER
from tecton_core.embeddings.model_artifacts import ModelArtifactProvider
from tecton_core.offline_store import DEFAULT_OPTIONS_PROVIDERS
from tecton_core.offline_store import OfflineStoreOptionsProvider
from tecton_core.query.data_sources import FileDataSourceScanNode
from tecton_core.query.data_sources import PushTableSourceScanNode
from tecton_core.query.data_sources import RedshiftDataSourceScanNode
from tecton_core.query.dialect import Dialect
from tecton_core.query.duckdb.rewrite import DuckDBTreeRewriter
from tecton_core.query.errors import UserDefinedTransformationError
from tecton_core.query.executor_params import ExecutionContext
from tecton_core.query.executor_params import QueryTreeStep
from tecton_core.query.executor_utils import DebugOutput
from tecton_core.query.executor_utils import QueryTreeMonitor
from tecton_core.query.executor_utils import get_stage_type_for_dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_utils import get_first_input_node_of_class
from tecton_core.query.node_utils import get_pipeline_dialect
from tecton_core.query.node_utils import get_staging_nodes
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.nodes import FeatureViewPipelineNode
from tecton_core.query.nodes import JoinNode
from tecton_core.query.nodes import MultiOdfvPipelineNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query.nodes import StagedTableScanNode
from tecton_core.query.nodes import StagingNode
from tecton_core.query.nodes import TextEmbeddingInferenceNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_core.query.optimize.adaptive import should_use_optimized_full_aggregate_node
from tecton_core.query.pandas.node import ArrowExecNode
from tecton_core.query.pandas.nodes import PandasDataSourceScanNode
from tecton_core.query.pandas.nodes import PandasFeatureViewPipelineNode
from tecton_core.query.pandas.nodes import PandasMultiOdfvPipelineNode
from tecton_core.query.pandas.nodes import PandasRenameColsNode
from tecton_core.query.pandas.nodes import PyArrowDataSourceScanNode
from tecton_core.query.query_tree_compute import ArrowCompute
from tecton_core.query.query_tree_compute import ComputeMonitor
from tecton_core.query.query_tree_compute import QueryTreeCompute
from tecton_core.query.query_tree_compute import SQLCompute
from tecton_core.secret_management import SecretResolver


logger = logging.getLogger(__name__)


def _pyarrow_type_contains_map_type(pyarrow_type: pyarrow.DataType) -> bool:
    if isinstance(pyarrow_type, pyarrow.MapType):
        return True
    elif isinstance(pyarrow_type, pyarrow.StructType):
        return any(_pyarrow_type_contains_map_type(field.type) for field in pyarrow_type)
    elif isinstance(pyarrow_type, pyarrow.ListType):
        return _pyarrow_type_contains_map_type(pyarrow_type.value_type)
    return False


@dataclass
class QueryTreeOutput:
    output: Optional[pyarrow.RecordBatchReader] = None

    @property
    def result_df(self) -> pandas.DataFrame:
        contains_map_type = any(_pyarrow_type_contains_map_type(field.type) for field in self.output.schema)
        if contains_map_type:
            # The `maps_as_pydicts` parameter for pyarrow.Table.to_pandas is only supported starting in pyarrow 13.0.0.
            if pyarrow.__version__ < "13.0.0":
                msg = f"Rift requires pyarrow>=13.0.0 to perform feature retrieval for Map features. You have version {pyarrow.__version__}."
                raise RuntimeError(msg)
            return self.output.read_pandas(maps_as_pydicts="strict")

        return self.output.read_pandas()

    @property
    def result_table(self) -> "pyarrow.RecordBatchReader":
        return self.output


UserErrors = (UserDefinedTransformationError,)


def get_data_source_dialect_and_physical_node(
    node: DataSourceScanNode,
) -> Tuple[Dialect, Optional[ArrowExecNode]]:
    batch_source = node.ds.batch_source
    if isinstance(batch_source, specs.PandasBatchSourceSpec):
        exec_node = PandasDataSourceScanNode.from_node_inputs(query_node=node, input_node=None)
        return Dialect.ARROW, exec_node
    elif isinstance(batch_source, specs.PyArrowBatchSourceSpec):
        exec_node = PyArrowDataSourceScanNode.from_node_inputs(query_node=node, input_node=None)
        return Dialect.ARROW, exec_node
    elif isinstance(batch_source, specs.FileSourceSpec):
        exec_node = FileDataSourceScanNode.from_node_input(node)
        return Dialect.ARROW, exec_node
    elif isinstance(
        batch_source,
        specs.PushTableSourceSpec,
    ):
        exec_node = PushTableSourceScanNode.from_node_input(node)
        return Dialect.ARROW, exec_node
    elif isinstance(batch_source, specs.SnowflakeSourceSpec):
        return Dialect.SNOWFLAKE, None
    elif isinstance(batch_source, specs.BigquerySourceSpec):
        return Dialect.BIGQUERY, None
    elif isinstance(batch_source, specs.RedshiftSourceSpec):
        return Dialect.ARROW, RedshiftDataSourceScanNode.from_node_input(node)

    msg = f"Unexpected data source type encountered: {batch_source.__class__}"
    raise Exception(msg)


def rewrite_data_sources(plan: NodeRef) -> None:
    """
    Inferring dialect for data source nodes and propagating it to the closest StagingNode.
    There are two kinds of data sources: sql-based and arrow-based.
    For arrow-based sources DataSourceScanNode is replaced with concrete implementation based on source type.
    """
    staging_nodes = get_staging_nodes(plan, QueryTreeStep.DATA_SOURCE, as_ref=True)
    if not staging_nodes:
        return

    for _, staging_node in staging_nodes.items():
        scan_node = get_first_input_node_of_class(staging_node, DataSourceScanNode)
        if not scan_node:
            continue

        dialect, physical_node = get_data_source_dialect_and_physical_node(scan_node)
        if physical_node:
            # replacing DataSourceScanNode
            staging_node.node = attrs.evolve(staging_node.node, dialect=dialect, input_node=physical_node.as_ref())
        else:
            # data source has SQL-based dialect
            # just need to propagate correct dialect to the StagingNode
            staging_node.node = attrs.evolve(staging_node.node, dialect=dialect)


def _rewrite_pandas_pipeline(plan: NodeRef) -> None:
    def traverse(tree: NodeRef) -> None:
        if isinstance(tree.node, FeatureViewPipelineNode):
            pipeline_node = tree.node
            # We don't need to rewrite inputs, because we assume that all inputs are StagingNodes
            assert all(
                isinstance(input_ref.node, StagingNode) for input_ref in pipeline_node.inputs_map.values()
            ), "All inputs to FeatureViewPipelineNode are expected to be StagingNode"

            physical_node = PandasFeatureViewPipelineNode.from_node_inputs(
                query_node=pipeline_node,
                input_node=None,
            )

            tree.node = StagingNode(
                dialect=Dialect.ARROW,
                compute_mode=ComputeMode.RIFT,
                input_node=physical_node.as_ref(),
                staging_table_name=f"pandas_pipeline_{pipeline_node.feature_definition_wrapper.name}",
            )
            return

        for i in tree.inputs:
            traverse(tree=i)

    traverse(plan)


def rewrite_pipeline_nodes(plan: NodeRef) -> None:
    """
    When pipeline mode is "pandas"/"python" the logical node is replaced with PandasFeatureViewPipelineNode + StagingNode.
    The latter is needed to indicate that this should be executed by ArrowCompute.

    In other cases (sql-based transformations): propagate correct dialect to the closest staging node.
    For now, all pipeline nodes are expected to have the same dialect.
    """
    pipeline_dialect = get_pipeline_dialect(plan)
    if pipeline_dialect == Dialect.PANDAS:
        _rewrite_pandas_pipeline(plan)
        return

    if not pipeline_dialect:
        return

    staging_nodes = get_staging_nodes(plan, QueryTreeStep.PIPELINE, as_ref=True)
    for _, staging_node in staging_nodes.items():
        staging_node.node = attrs.evolve(staging_node.node, dialect=pipeline_dialect)


def rewrite_rtfvs(plan: NodeRef) -> None:
    """
    Logical node MultiOdfvPipelineNode is replaced with PandasMultiOdfvPipelineNode,
    which handles both "pandas" and "python" modes.

    If MultiOdfvPipelineNode is succeeded by RenameColsNode it should as well replaced by PandasRenameColsNode
    to minimize switching between Arrow and DuckDB computes.
    """

    def create_physical_odfv_pipeline_node(logical_node: MultiOdfvPipelineNode) -> ArrowExecNode:
        return PandasMultiOdfvPipelineNode.from_node_inputs(logical_node, logical_node.input_node)

    def create_physical_rename_node(logical_node: RenameColsNode, odfv_pipeline_node: NodeRef) -> ArrowExecNode:
        return PandasRenameColsNode.from_node_inputs(logical_node, odfv_pipeline_node)

    def traverse(tree: NodeRef) -> None:
        if isinstance(tree.node, RenameColsNode) and isinstance(tree.node.input_node.node, MultiOdfvPipelineNode):
            tree.node = StagingNode(
                dialect=Dialect.ARROW,
                compute_mode=ComputeMode.RIFT,
                input_node=create_physical_rename_node(
                    tree.node, create_physical_odfv_pipeline_node(tree.node.input_node.node).as_ref()
                ).as_ref(),
                staging_table_name="rtfv_output",
            )
            return

        if isinstance(tree.node, MultiOdfvPipelineNode):
            tree.node = StagingNode(
                dialect=Dialect.ARROW,
                compute_mode=ComputeMode.RIFT,
                input_node=create_physical_odfv_pipeline_node(tree.node).as_ref(),
                staging_table_name="rtfv_output",
            )
            return

        for i in tree.inputs:
            traverse(tree=i)

    traverse(plan)


def rewrite_user_input(plan: NodeRef) -> None:
    """
    UserSpecifiedDataNode is wrapped into StagingNode with Arrow dialect
    to force executor use ArrowCompute and call .to_arrow_reader() instead of .to_sql()
    """

    def traverse(tree: NodeRef) -> None:
        if isinstance(tree.node, UserSpecifiedDataNode):
            tree.node = StagingNode(
                dialect=Dialect.ARROW,
                compute_mode=ComputeMode.RIFT,
                input_node=tree.node.as_ref(),
                staging_table_name=tree.node.data._temp_table_name,
            )
            return

        for i in tree.inputs:
            traverse(tree=i)

    traverse(plan)


def rewrite_embedding_nodes(plan: NodeRef) -> None:
    """
    UserSpecifiedDataNode is wrapped into StagingNode with Arrow dialect
    to force executor use ArrowCompute and call .to_arrow_reader() instead of .to_sql()
    """

    def traverse(tree: NodeRef) -> None:
        if isinstance(tree.node, TextEmbeddingInferenceNode):
            from tecton_core.embeddings.nodes import ArrowExecTextEmbeddingInferenceNode

            tree.node = ArrowExecTextEmbeddingInferenceNode.from_node_input(tree.node)
            return

        for i in tree.inputs:
            traverse(tree=i)

    traverse(plan)


def rewrite_cross_feature_views_joins(plan: NodeRef) -> None:
    """
    Inserts StagingNode with checkpoint=True:
    1) after each feature view's aggregations are calculated and before join with other feature views
    2) in between every N join
        eg, when N is 3 and number of batch feature views is 12, there will be 3 additional checkpoints
        (there should be no checkpoints at the top of the QT)
    """
    if not conf.DUCKDB_ENABLE_CHECKPOINTING.enabled():
        return

    checkpoint_every_n_joins = int(conf.get_or_raise("DUCKDB_CHECKPOINT_EVERY_N_JOIN"))

    def traverse(node_ref: NodeRef, depth: int = 1) -> None:
        if not isinstance(node_ref.node, JoinNode):
            for i in node_ref.inputs:
                traverse(i, depth)
            return

        join_node = node_ref.node
        assert isinstance(join_node, JoinNode)

        # right side of join is next feature view retrieval
        right_branch = join_node.right
        # always staging results of right side
        right_branch = StagingNode(
            right_branch.node.dialect,
            right_branch.node.compute_mode,
            right_branch,
            staging_table_name=f"checkpoint_feature_view_{depth}",
            checkpoint=True,
        ).as_ref()

        # left side of join is the result of previous joins
        left_branch = join_node.left
        if isinstance(left_branch.node, JoinNode) and depth and depth % checkpoint_every_n_joins == 0:
            left_branch = StagingNode(
                left_branch.node.dialect,
                left_branch.node.compute_mode,
                left_branch,
                staging_table_name=f"checkpoint_join_output_{depth}",
                checkpoint=True,
            ).as_ref()

        node_ref.node = attrs.evolve(node_ref.node, right=right_branch, left=left_branch)

        traverse(left_branch, depth=depth + 1)

    traverse(plan)


def logical_plan_to_physical_plan(logical_plan: NodeRef, use_optimized_full_agg: bool = False) -> NodeRef:
    physical_plan = logical_plan.deepcopy()

    # replace some generic nodes with Rift specific
    # ToDo: this can be removed when Athena and Snowflake are removed
    # and the code can be moved to generic nodes
    rewriter = DuckDBTreeRewriter()
    rewriter.rewrite(physical_plan, use_optimized_full_agg)

    rewrite_data_sources(physical_plan)
    rewrite_pipeline_nodes(physical_plan)
    rewrite_rtfvs(physical_plan)
    rewrite_user_input(physical_plan)
    rewrite_embedding_nodes(physical_plan)
    rewrite_cross_feature_views_joins(physical_plan)

    return physical_plan


@attrs.define
class Stage:
    """
    Stage is subtree of a physical plan, where all nodes have the same dialect and thus can be executed in one go
    """

    dialect: Dialect
    output: NodeRef
    inputs: List[NodeRef]
    description: Optional[str] = None


def split_plan_into_stages(plan: NodeRef) -> List[List[Stage]]:
    """
    Split the plan into stages using StagingNodes as splitting points.
    Inputs in each stage, which are StagingNodes in the original plan, are replaced with StagedTableScanNode.

    Run breadth-first traverse over the plan to also split stages into levels. Such that a stage on certain level
    cannot be executed until all stages on the lower level are completed, since they can be inputs to the current stage.
    Stages on the same level, however, can be executed in parallel.

    :return: stages grouped by levels
    """

    def find_inputs(tree: NodeRef, output: StagingNode) -> Iterable[NodeRef]:
        if (
            isinstance(tree.node, StagingNode)
            and tree.node != output
            and (tree.node.dialect != output.dialect or tree.node.checkpoint)
        ):
            # Only staging nodes of different dialect count as input to the current stage
            # or we encountered a staging node with "checkpoint" enabled
            yield tree
            return

        for input_ in tree.inputs:
            yield from find_inputs(input_, output)

    def traverse():
        while True:
            current_level = levels[-1]
            next_level = []

            for stage in current_level:
                for input_ in stage.inputs:
                    next_stage_output = input_.node
                    assert isinstance(next_stage_output, StagingNode)
                    assert next_stage_output.dialect, f"Dialect must be set on StagingNode: {next_stage_output}"
                    next_level.append(
                        Stage(
                            dialect=next_stage_output.dialect,
                            output=next_stage_output.as_ref(),
                            inputs=list(find_inputs(next_stage_output.as_ref(), next_stage_output)),
                            description=next_stage_output.stage_description,
                        )
                    )

                    input_.node = StagedTableScanNode(
                        input_.node.dialect,
                        input_.node.compute_mode,
                        staged_schema=input_.node.output_schema,
                        staging_table_name=input_.node.staging_table_name_unique(),
                    )
            if not next_level:
                return

            levels.append(next_level)

    if not isinstance(plan.node, StagingNode):
        # for simplicity plan should always have StagingNode at the root
        plan = StagingNode(
            dialect=Dialect.DUCKDB,
            compute_mode=ComputeMode.RIFT,
            input_node=plan,
            staging_table_name="",
        ).as_ref()

    levels = [
        [
            Stage(
                dialect=plan.node.dialect,
                output=plan,
                inputs=list(find_inputs(plan, plan.node)),
                description=plan.node.stage_description,
            )
        ]
    ]
    traverse()
    return levels


@attrs.define
class QueryTreeExecutor:
    offline_store_options_providers: Iterable[OfflineStoreOptionsProvider] = DEFAULT_OPTIONS_PROVIDERS
    secret_resolver: Optional[SecretResolver] = None
    model_artifact_provider: Optional[ModelArtifactProvider] = DEFAULT_MODEL_PROVIDER
    monitor: QueryTreeMonitor = DebugOutput()
    is_debug: bool = attrs.field(init=False)
    # TODO: Put duckdb_config in a map when we have more configs for different dialects.
    duckdb_config: Optional[DuckDBConfig] = None
    _dialect_to_compute_map: Dict[Dialect, QueryTreeCompute] = {}
    # Used to track temp tables per dialect so we can clean them up appropriately & avoid re-registering duplicates
    _dialect_to_temp_table_name: Optional[Dict[Dialect, set]] = attrs.field(init=False)

    def __attrs_post_init__(self):
        # TODO(danny): Expose as configs
        self.is_debug = conf.get_bool("DUCKDB_DEBUG")
        self._dialect_to_temp_table_name = None

    @contextlib.contextmanager
    def _monitor_stage(
        self,
        step: str,
        type_: Optional[int] = None,
        dialect: Optional[Dialect] = None,
    ) -> Iterator[ComputeMonitor]:
        assert type_ or dialect, "Either type or dialect must be provided"
        if not type_:
            type_ = get_stage_type_for_dialect(dialect)

        monitor_stage_id = self.monitor.create_stage(type_, step)

        try:
            self.monitor.update_progress(monitor_stage_id, 0)
            yield ComputeMonitor(
                log_progress=lambda p: self.monitor.update_progress(monitor_stage_id, p),
                set_query=lambda q: self.monitor.set_query(monitor_stage_id, q),
            )
        except UserErrors:
            self.monitor.set_failed(monitor_stage_id, user_error=True)
            raise
        except Exception:
            self.monitor.set_failed(monitor_stage_id, user_error=False)
            raise
        else:
            self.monitor.update_progress(monitor_stage_id, 1)
            self.monitor.set_completed(monitor_stage_id)

    def exec_qt(self, logical_plan: NodeRef) -> QueryTreeOutput:
        # Make copy so the execution doesn't mutate the original QT visible to users

        physical_plan = logical_plan_to_physical_plan(
            logical_plan,
            use_optimized_full_agg=should_use_optimized_full_aggregate_node(logical_plan),
        )
        if self.is_debug:
            logger.warning("---------------------------------- Executing overall QT ----------------------------------")
            logger.warning(f"QT: \n{logical_plan.pretty_str()}")
            logger.warning("---------------------------------- Physical plan -----------------------------------------")
            logger.warning(f"QT: \n{physical_plan.pretty_str()}")

        stage_levels = split_plan_into_stages(physical_plan)
        inputs = {}

        # Processing levels from bottom to top
        for stages in reversed(stage_levels):
            inputs = self._process_staging_nodes(stages, inputs)

        return QueryTreeOutput(output=next(iter(inputs.values())))

    def _process_staging_nodes(
        self,
        stages: List[Stage],
        inputs: Dict[str, pyarrow.RecordBatchReader],
    ) -> Dict[str, pyarrow.RecordBatchReader]:
        readers = {}
        for stage in stages:
            assert isinstance(stage.output.node, StagingNode)
            compute = self._get_or_create_compute_by_dialect(stage.dialect, stage.output)

            if self.is_debug:
                logger.warning("---------------------------------- Executing stage ----------------------------------")
                logger.warning(f"QT: \n{stage.output.pretty_str()}")

            input_names = {node_ref.node.staging_table_name for node_ref in stage.inputs}
            stage_inputs = {table: reader for table, reader in inputs.items() if table in input_names}

            with compute:
                name, reader = self._process_staging_node(
                    stage.output.node,
                    inputs=stage_inputs,
                    compute=compute,
                    monitor=None,
                )
            readers[name] = reader
        return readers

    def _process_staging_node(
        self,
        staging_node: StagingNode,
        inputs: Dict[str, pyarrow.RecordBatchReader],
        compute: QueryTreeCompute,
        monitor: ComputeMonitor,
    ) -> Tuple[str, pyarrow.RecordBatchReader]:
        start_time = datetime.now()
        staging_table_name = staging_node.staging_table_name_unique()

        if isinstance(compute, ArrowCompute):
            context = ExecutionContext(
                offline_store_options_providers=self.offline_store_options_providers,
                secret_resolver=self.secret_resolver,
                model_artifact_provider=self.model_artifact_provider,
            )
            reader = compute.run(staging_node.as_ref(), inputs, context, monitor=monitor)
            return staging_table_name, reader

        assert isinstance(compute, SQLCompute)
        for table_name, pa_reader in inputs.items():
            if pa_reader:
                compute.register_temp_table(table_name, pa_reader)

        sql_string = staging_node.with_dialect(compute.get_dialect())._to_staging_query_sql()
        expected_output_schema = staging_node.output_schema if len(staging_node.output_schema) else None
        checkpoint_as = staging_node.staging_table_name_unique() if staging_node.checkpoint else None
        return_dataframe = True if not checkpoint_as else False

        reader = compute.run_sql(
            sql_string,
            return_dataframe=return_dataframe,
            expected_output_schema=expected_output_schema,
            monitor=monitor,
            checkpoint_as=checkpoint_as,
        )

        # Cleaning up checkpointed tables
        checkpointed_inputs = [table for table, reader in inputs.items() if not reader]
        for table_name in checkpointed_inputs:
            compute.unregister_temp_table(table_name)

        staging_done_time = datetime.now()
        if self.is_debug:
            elapsed_staging_time = (staging_done_time - start_time).total_seconds()
            logger.warning(f"STAGE_{staging_table_name}_TIME_SEC: {elapsed_staging_time}")

        return staging_table_name, reader

    def _get_or_create_compute_by_dialect(
        self,
        dialect: Dialect,
        qt_root: Optional[NodeRef] = None,
    ) -> QueryTreeCompute:
        if dialect in self._dialect_to_compute_map:
            return self._dialect_to_compute_map[dialect]

        compute = QueryTreeCompute.for_dialect(dialect, self, qt_root)
        self._dialect_to_compute_map[dialect] = compute
        return compute



from tecton_proto.common import schema__client_pb2 as _schema__client_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SchemaContainer(_message.Message):
    __slots__ = ["tecton_schema"]
    TECTON_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    tecton_schema: _schema__client_pb2.Schema
    def __init__(self, tecton_schema: _Optional[_Union[_schema__client_pb2.Schema, _Mapping]] = ...) -> None: ...



from tecton_proto.realtime import instance_group__client_pb2 as _instance_group__client_pb2
from tecton_proto.server_groups import server_group_states__client_pb2 as _server_group_states__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ServerGroupClusterConfig(_message.Message):
    __slots__ = ["transform_server_group"]
    TRANSFORM_SERVER_GROUP_FIELD_NUMBER: _ClassVar[int]
    transform_server_group: _containers.RepeatedCompositeFieldContainer[TransformServerGroupInfo]
    def __init__(self, transform_server_group: _Optional[_Iterable[_Union[TransformServerGroupInfo, _Mapping]]] = ...) -> None: ...

class TransformServerGroupInfo(_message.Message):
    __slots__ = ["autoscaling_policy", "aws_instance_group_config", "environment_variables", "server_group_id"]
    class EnvironmentVariablesEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    AUTOSCALING_POLICY_FIELD_NUMBER: _ClassVar[int]
    AWS_INSTANCE_GROUP_CONFIG_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_VARIABLES_FIELD_NUMBER: _ClassVar[int]
    SERVER_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    autoscaling_policy: _server_group_states__client_pb2.AutoscalingPolicy
    aws_instance_group_config: _instance_group__client_pb2.AWSInstanceGroupUpdateConfig
    environment_variables: _containers.ScalarMap[str, str]
    server_group_id: str
    def __init__(self, server_group_id: _Optional[str] = ..., autoscaling_policy: _Optional[_Union[_server_group_states__client_pb2.AutoscalingPolicy, _Mapping]] = ..., aws_instance_group_config: _Optional[_Union[_instance_group__client_pb2.AWSInstanceGroupUpdateConfig, _Mapping]] = ..., environment_variables: _Optional[_Mapping[str, str]] = ...) -> None: ...



from tecton_proto.common import id__client_pb2 as _id__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Acl(_message.Message):
    __slots__ = ["api_key_ids", "api_key_str_ids", "okta_ids", "permission"]
    API_KEY_IDS_FIELD_NUMBER: _ClassVar[int]
    API_KEY_STR_IDS_FIELD_NUMBER: _ClassVar[int]
    OKTA_IDS_FIELD_NUMBER: _ClassVar[int]
    PERMISSION_FIELD_NUMBER: _ClassVar[int]
    api_key_ids: _containers.RepeatedCompositeFieldContainer[_id__client_pb2.Id]
    api_key_str_ids: _containers.RepeatedScalarFieldContainer[str]
    okta_ids: _containers.RepeatedScalarFieldContainer[str]
    permission: str
    def __init__(self, permission: _Optional[str] = ..., api_key_ids: _Optional[_Iterable[_Union[_id__client_pb2.Id, _Mapping]]] = ..., okta_ids: _Optional[_Iterable[str]] = ..., api_key_str_ids: _Optional[_Iterable[str]] = ...) -> None: ...



from tecton_proto.args import transformation__client_pb2 as _transformation__client_pb2
from tecton_proto.args import user_defined_function__client_pb2 as _user_defined_function__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.data import fco_metadata__client_pb2 as _fco_metadata__client_pb2
from tecton_proto.validation import validator__client_pb2 as _validator__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Transformation(_message.Message):
    __slots__ = ["fco_metadata", "options", "transformation_id", "transformation_mode", "user_function", "validation_args"]
    class OptionsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    FCO_METADATA_FIELD_NUMBER: _ClassVar[int]
    OPTIONS_FIELD_NUMBER: _ClassVar[int]
    TRANSFORMATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRANSFORMATION_MODE_FIELD_NUMBER: _ClassVar[int]
    USER_FUNCTION_FIELD_NUMBER: _ClassVar[int]
    VALIDATION_ARGS_FIELD_NUMBER: _ClassVar[int]
    fco_metadata: _fco_metadata__client_pb2.FcoMetadata
    options: _containers.ScalarMap[str, str]
    transformation_id: _id__client_pb2.Id
    transformation_mode: _transformation__client_pb2.TransformationMode
    user_function: _user_defined_function__client_pb2.UserDefinedFunction
    validation_args: _validator__client_pb2.TransformationValidationArgs
    def __init__(self, transformation_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., fco_metadata: _Optional[_Union[_fco_metadata__client_pb2.FcoMetadata, _Mapping]] = ..., transformation_mode: _Optional[_Union[_transformation__client_pb2.TransformationMode, str]] = ..., user_function: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., validation_args: _Optional[_Union[_validator__client_pb2.TransformationValidationArgs, _Mapping]] = ..., options: _Optional[_Mapping[str, str]] = ...) -> None: ...



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ArrayValue(_message.Message):
    __slots__ = ["array_values", "bool_values", "float32_values", "float64_values", "int64_values", "map_values", "null_indices", "string_values", "struct_values"]
    ARRAY_VALUES_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUES_FIELD_NUMBER: _ClassVar[int]
    FLOAT32_VALUES_FIELD_NUMBER: _ClassVar[int]
    FLOAT64_VALUES_FIELD_NUMBER: _ClassVar[int]
    INT64_VALUES_FIELD_NUMBER: _ClassVar[int]
    MAP_VALUES_FIELD_NUMBER: _ClassVar[int]
    NULL_INDICES_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUES_FIELD_NUMBER: _ClassVar[int]
    STRUCT_VALUES_FIELD_NUMBER: _ClassVar[int]
    array_values: _containers.RepeatedCompositeFieldContainer[ArrayValue]
    bool_values: _containers.RepeatedScalarFieldContainer[bool]
    float32_values: _containers.RepeatedScalarFieldContainer[float]
    float64_values: _containers.RepeatedScalarFieldContainer[float]
    int64_values: _containers.RepeatedScalarFieldContainer[int]
    map_values: _containers.RepeatedCompositeFieldContainer[MapValue]
    null_indices: _containers.RepeatedScalarFieldContainer[int]
    string_values: _containers.RepeatedScalarFieldContainer[str]
    struct_values: _containers.RepeatedCompositeFieldContainer[StructValue]
    def __init__(self, string_values: _Optional[_Iterable[str]] = ..., array_values: _Optional[_Iterable[_Union[ArrayValue, _Mapping]]] = ..., struct_values: _Optional[_Iterable[_Union[StructValue, _Mapping]]] = ..., map_values: _Optional[_Iterable[_Union[MapValue, _Mapping]]] = ..., bool_values: _Optional[_Iterable[bool]] = ..., int64_values: _Optional[_Iterable[int]] = ..., float32_values: _Optional[_Iterable[float]] = ..., float64_values: _Optional[_Iterable[float]] = ..., null_indices: _Optional[_Iterable[int]] = ...) -> None: ...

class BatchCompactedDataRow(_message.Message):
    __slots__ = ["compacted_tile"]
    COMPACTED_TILE_FIELD_NUMBER: _ClassVar[int]
    compacted_tile: FeatureValueList
    def __init__(self, compacted_tile: _Optional[_Union[FeatureValueList, _Mapping]] = ...) -> None: ...

class BatchCompactedDataRowV2(_message.Message):
    __slots__ = ["compacted_tiles"]
    COMPACTED_TILES_FIELD_NUMBER: _ClassVar[int]
    compacted_tiles: _containers.RepeatedCompositeFieldContainer[FeatureValueList]
    def __init__(self, compacted_tiles: _Optional[_Iterable[_Union[FeatureValueList, _Mapping]]] = ...) -> None: ...

class CachedFeatureView(_message.Message):
    __slots__ = ["effective_time", "feature_statuses", "feature_values", "updated_at"]
    EFFECTIVE_TIME_FIELD_NUMBER: _ClassVar[int]
    FEATURE_STATUSES_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VALUES_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    effective_time: _timestamp_pb2.Timestamp
    feature_statuses: _containers.RepeatedScalarFieldContainer[int]
    feature_values: _containers.RepeatedCompositeFieldContainer[FeatureValue]
    updated_at: _timestamp_pb2.Timestamp
    def __init__(self, feature_values: _Optional[_Iterable[_Union[FeatureValue, _Mapping]]] = ..., feature_statuses: _Optional[_Iterable[int]] = ..., effective_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...) -> None: ...

class FeatureValue(_message.Message):
    __slots__ = ["array_value", "bool_value", "float32_value", "float64_value", "int64_value", "map_value", "null_value", "string_value", "struct_value"]
    ARRAY_VALUE_FIELD_NUMBER: _ClassVar[int]
    BOOL_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT32_VALUE_FIELD_NUMBER: _ClassVar[int]
    FLOAT64_VALUE_FIELD_NUMBER: _ClassVar[int]
    INT64_VALUE_FIELD_NUMBER: _ClassVar[int]
    MAP_VALUE_FIELD_NUMBER: _ClassVar[int]
    NULL_VALUE_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    STRUCT_VALUE_FIELD_NUMBER: _ClassVar[int]
    array_value: ArrayValue
    bool_value: bool
    float32_value: float
    float64_value: float
    int64_value: int
    map_value: MapValue
    null_value: NullValue
    string_value: str
    struct_value: StructValue
    def __init__(self, array_value: _Optional[_Union[ArrayValue, _Mapping]] = ..., float64_value: _Optional[float] = ..., float32_value: _Optional[float] = ..., int64_value: _Optional[int] = ..., bool_value: bool = ..., string_value: _Optional[str] = ..., null_value: _Optional[_Union[NullValue, _Mapping]] = ..., struct_value: _Optional[_Union[StructValue, _Mapping]] = ..., map_value: _Optional[_Union[MapValue, _Mapping]] = ...) -> None: ...

class FeatureValueList(_message.Message):
    __slots__ = ["feature_values"]
    FEATURE_VALUES_FIELD_NUMBER: _ClassVar[int]
    feature_values: _containers.RepeatedCompositeFieldContainer[FeatureValue]
    def __init__(self, feature_values: _Optional[_Iterable[_Union[FeatureValue, _Mapping]]] = ...) -> None: ...

class MapValue(_message.Message):
    __slots__ = ["keys", "values"]
    KEYS_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    keys: ArrayValue
    values: ArrayValue
    def __init__(self, keys: _Optional[_Union[ArrayValue, _Mapping]] = ..., values: _Optional[_Union[ArrayValue, _Mapping]] = ...) -> None: ...

class NullValue(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class RedisTAFVData(_message.Message):
    __slots__ = ["anchor_time", "feature_values", "written_by_batch"]
    ANCHOR_TIME_FIELD_NUMBER: _ClassVar[int]
    FEATURE_VALUES_FIELD_NUMBER: _ClassVar[int]
    WRITTEN_BY_BATCH_FIELD_NUMBER: _ClassVar[int]
    anchor_time: int
    feature_values: _containers.RepeatedCompositeFieldContainer[FeatureValue]
    written_by_batch: bool
    def __init__(self, anchor_time: _Optional[int] = ..., written_by_batch: bool = ..., feature_values: _Optional[_Iterable[_Union[FeatureValue, _Mapping]]] = ...) -> None: ...

class StructValue(_message.Message):
    __slots__ = ["values"]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    values: _containers.RepeatedCompositeFieldContainer[FeatureValue]
    def __init__(self, values: _Optional[_Iterable[_Union[FeatureValue, _Mapping]]] = ...) -> None: ...



BUILD_STATUS = {'BUILD_ID': 'uncommitted-buildkite-agent-1737474387', 'BUILD_TIMESTAMP': '1737474387', 'FORMATTED_DATE': '2025 Jan 21 15 46 27 Tue', 'GIT_COMMIT': '24fa54c5c27a37908ac40133d19e9735584fabc8', 'BUILD_EMBED_LABEL': '', 'BUILD_HOST': 'buildkite-agent-medium-8564564c5c-dv8gr', 'BUILD_USER': 'buildkite-agent'}



from tecton_proto.common import id__client_pb2 as _id__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class FeatureRepoSourceInfo(_message.Message):
    __slots__ = ["source_info"]
    SOURCE_INFO_FIELD_NUMBER: _ClassVar[int]
    source_info: _containers.RepeatedCompositeFieldContainer[SourceInfo]
    def __init__(self, source_info: _Optional[_Iterable[_Union[SourceInfo, _Mapping]]] = ...) -> None: ...

class SourceInfo(_message.Message):
    __slots__ = ["fco_id", "scope", "source_filename", "source_lineno"]
    FCO_ID_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FILENAME_FIELD_NUMBER: _ClassVar[int]
    SOURCE_LINENO_FIELD_NUMBER: _ClassVar[int]
    fco_id: _id__client_pb2.Id
    scope: str
    source_filename: str
    source_lineno: str
    def __init__(self, fco_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., scope: _Optional[str] = ..., source_lineno: _Optional[str] = ..., source_filename: _Optional[str] = ...) -> None: ...



import enum
import logging

from tecton_core.errors import INTERNAL_ERROR_FROM_MDS
from tecton_core.errors import FailedPreconditionError
from tecton_core.errors import TectonAbortedError
from tecton_core.errors import TectonAlreadyExistsError
from tecton_core.errors import TectonAPIInaccessibleError
from tecton_core.errors import TectonAPIValidationError
from tecton_core.errors import TectonDeadlineExceededError
from tecton_core.errors import TectonNotFoundError
from tecton_core.errors import TectonNotImplementedError
from tecton_core.errors import TectonOperationCancelledError
from tecton_core.errors import TectonResourceExhaustedError
from tecton_core.metadata_service_impl import trace
from tecton_core.metadata_service_impl.providers import RequestProvider


logger = logging.getLogger(__name__)


class gRPCStatus(enum.Enum):
    """gRPC response status codes.

    Status codes are replicated here to avoid importing the `grpc.StatusCode` enum class,
    which requires the grpcio library.

    https://grpc.github.io/grpc/core/md_doc_statuscodes.html
    """

    OK = 0
    CANCELLED = 1
    UNKNOWN = 2
    INVALID_ARGUMENT = 3
    DEADLINE_EXCEEDED = 4
    NOT_FOUND = 5
    ALREADY_EXISTS = 6
    PERMISSION_DENIED = 7
    RESOURCE_EXHAUSTED = 8
    FAILED_PRECONDITION = 9
    ABORTED = 10
    OUT_OF_RANGE = 11
    UNIMPLEMENTED = 12
    INTERNAL = 13
    UNAVAILABLE = 14
    DATA_LOSS = 15
    UNAUTHENTICATED = 16


def raise_for_grpc_status(status_code: int, details: str, host_url: str, request_provider: RequestProvider) -> None:
    """
    Raise an exception based on a gRPC error status code.
    """

    if status_code == gRPCStatus.OK.value:
        return

    # Error handling
    if status_code == gRPCStatus.UNAVAILABLE.value:
        raise TectonAPIInaccessibleError(details, host_url)

    if status_code == gRPCStatus.INVALID_ARGUMENT.value:
        raise TectonAPIValidationError(details)

    if status_code == gRPCStatus.FAILED_PRECONDITION.value:
        raise FailedPreconditionError(details)

    if status_code == gRPCStatus.UNAUTHENTICATED.value:
        msg = f"Tecton credentials are invalid, not configured, or expired ({details}). To authenticate using an API key, set TECTON_API_KEY in your environment or use tecton.login(tecton_api_key=<key>, tecton_url=<url>). To authenticate as your user, run `tecton login` with the CLI or `tecton.login(tecton_url=<url>)` in your notebook."
        raise PermissionError(msg)

    if status_code == gRPCStatus.PERMISSION_DENIED.value:
        msg = f"Insufficient permissions ({details})."
        raise PermissionError(msg)

    if status_code == gRPCStatus.NOT_FOUND.value:
        raise TectonNotFoundError(details)

    if status_code == gRPCStatus.UNIMPLEMENTED.value:
        raise TectonNotImplementedError(details)

    if status_code == gRPCStatus.CANCELLED.value:
        raise TectonOperationCancelledError(details)

    if status_code == gRPCStatus.DEADLINE_EXCEEDED.value:
        raise TectonDeadlineExceededError(details)

    if status_code == gRPCStatus.ALREADY_EXISTS.value:
        raise TectonAlreadyExistsError(details)

    if status_code == gRPCStatus.RESOURCE_EXHAUSTED.value:
        raise TectonResourceExhaustedError(details)

    if status_code == gRPCStatus.ABORTED.value:
        raise TectonAbortedError(details)

    if status_code == gRPCStatus.OUT_OF_RANGE.value:
        raise TectonAPIValidationError(details)

    logger.debug(f"Unknown MDS exception. code={status_code}, details={details}")

    raise INTERNAL_ERROR_FROM_MDS(details, trace.get_trace_id())



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor
SERVER_GROUP_STATUS_CREATING: ServerGroupStatus
SERVER_GROUP_STATUS_DELETING: ServerGroupStatus
SERVER_GROUP_STATUS_ERROR: ServerGroupStatus
SERVER_GROUP_STATUS_PENDING: ServerGroupStatus
SERVER_GROUP_STATUS_READY: ServerGroupStatus
SERVER_GROUP_STATUS_UNSPECIFIED: ServerGroupStatus
SERVER_GROUP_STATUS_UPDATING: ServerGroupStatus

class ServerGroupStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from datetime import timedelta
from functools import reduce
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import attrs
import pyspark
import pyspark.sql.window as spark_window
from pyspark.sql import functions as F
from pyspark.sql.functions import expr
from pyspark.sql.types import ArrayType
from pyspark.sql.types import LongType

import tecton_core.tecton_pendulum as pendulum
from tecton_core import time_utils as core_time_utils
from tecton_core.compute_mode import ComputeMode
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.feature_extraction_sql_builder import FeatureExtractionSqlBuilder
from tecton_core.query_consts import TECTON_TEMP_AGGREGATION_SECONDARY_KEY_COL
from tecton_core.query_consts import anchor_time
from tecton_core.query_consts import valid_from
from tecton_core.query_consts import valid_to
from tecton_core.specs import MaterializedFeatureViewSpec
from tecton_core.specs import MaterializedFeatureViewType
from tecton_proto.common.aggregation_function__client_pb2 import AggregationFunction
from tecton_proto.data import feature_view__client_pb2 as feature_view__data_pb2
from tecton_spark.query.node import SparkExecNode
from tecton_spark.time_utils import convert_epoch_to_timestamp_column
from tecton_spark.time_utils import convert_timestamp_to_epoch


@attrs.frozen
class AddAnchorTimeSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_store_format_version: int
    batch_schedule: int
    timestamp_field: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        anchor_time_val = convert_timestamp_to_epoch(F.col(self.timestamp_field), self.feature_store_format_version)
        df = input_df.withColumn(
            anchor_time(),
            anchor_time_val - anchor_time_val % self.batch_schedule,
        )
        return df


@attrs.frozen
class AddRetrievalAnchorTimeSparkNode(SparkExecNode):
    input_node: SparkExecNode
    name: str
    feature_store_format_version: int
    batch_schedule: int
    tile_interval: int
    timestamp_field: str
    is_stream: bool
    data_delay_seconds: Optional[int] = 0

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        anchor_time_val = convert_timestamp_to_epoch(
            F.col(self.timestamp_field) - expr(f"interval {self.data_delay_seconds} seconds"),
            self.feature_store_format_version,
        )
        # tile_interval will be 0 for continuous
        if self.tile_interval == 0:
            df = input_df.withColumn(anchor_time(), anchor_time_val)
        else:
            # For stream, we use the tile interval for bucketing since the data is available as soon as
            # the aggregation interval ends.
            # For BAFV, we use the batch schedule to get the last tile written.
            if self.is_stream:
                df = input_df.withColumn(
                    anchor_time(),
                    anchor_time_val - anchor_time_val % self.tile_interval - self.tile_interval,
                )
            else:
                df = input_df.withColumn(
                    anchor_time(),
                    anchor_time_val - anchor_time_val % self.batch_schedule - self.tile_interval,
                )
        return df


@attrs.frozen
class MultiRtfvFeatureExtractionSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_definition_namespaces: List[Tuple[FeatureDefinitionWrapper, str]]
    events_df_timestamp_field: str
    use_namespace_feature_prefix: bool = True

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)

        sql_builder = FeatureExtractionSqlBuilder(
            self.events_df_timestamp_field,
            self.input_node.columns,
            ComputeMode.SPARK,
            self.use_namespace_feature_prefix,
        )

        feature_query_terms = []
        for fdw, namespace in self.feature_definition_namespaces:
            feature_query_terms.extend(sql_builder.extract_rtfv_features(fdw, namespace))

        # quote_char="`" is necessary to ensure column references show up as `select `column_name`` instead of `select 'column_name'`.
        feature_sql_exprs = [expr(col.get_sql(with_alias=True, quote_char="`")) for col in feature_query_terms]
        return input_df.select("*", *feature_sql_exprs)


@attrs.frozen
class RenameColsSparkNode(SparkExecNode):
    input_node: SparkExecNode
    mapping: Optional[Dict[str, Union[str, List[str]]]]
    drop: Optional[List[str]]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        if self.mapping:
            rename_cols = []
            for old_name, new_name in self.mapping.items():
                if not new_name:
                    continue
                if isinstance(new_name, list):
                    for col in sorted(new_name):
                        rename_cols.append(F.col(old_name).alias(col))
                else:
                    rename_cols.append(F.col(old_name).alias(new_name))

            existing_cols = [col for col in input_df.columns if col not in self.mapping.keys()]
            input_df = input_df.select(*existing_cols, *rename_cols)

        if self.drop:
            input_df = input_df.drop(*self.drop)

        return input_df


@attrs.frozen
class ConvertEpochToTimestampSparkNode(SparkExecNode):
    input_node: SparkExecNode
    feature_store_formats: Dict[str, int]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        for name, feature_store_format_version in self.feature_store_formats.items():
            input_df = input_df.withColumn(
                name,
                convert_epoch_to_timestamp_column(F.col(name), feature_store_format_version),
            )
        return input_df


@attrs.frozen
class ConvertTimestampToUTCSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_key: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return self.input_node.to_dataframe(spark)


@attrs.frozen
class AddAnchorTimeColumnsForSawtoothIntervalsSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_field: str
    anchor_time_column_map: Dict[str, timedelta]
    data_delay_seconds: int
    feature_store_format_version: int
    aggregation_tile_interval_column_map: Dict[str, int]
    truncate_to_recent_complete_tile: bool

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        if self.timestamp_field != anchor_time():
            anchor_time_val = convert_timestamp_to_epoch(
                F.col(self.timestamp_field) - expr(f"interval {self.data_delay_seconds} seconds"),
                self.feature_store_format_version,
            )
        else:
            anchor_time_val = F.col(self.timestamp_field)

        columns = [F.col(column) for column in self.input_node.columns]
        new_timestamp_columns = []
        for column_name, unit in self.anchor_time_column_map.items():
            interval_seconds = core_time_utils.convert_timedelta_for_version(unit, self.feature_store_format_version)
            if interval_seconds == 0:
                spark_column = anchor_time_val.alias(column_name)
            else:
                aggregation_tile_interval = (
                    self.aggregation_tile_interval_column_map[column_name]
                    if self.truncate_to_recent_complete_tile
                    else 0
                )
                spark_column = (anchor_time_val - anchor_time_val % interval_seconds - aggregation_tile_interval).alias(
                    column_name
                )
            new_timestamp_columns.append(spark_column)
        return input_df.select(*columns, *new_timestamp_columns)


@attrs.frozen
class AdjustAnchorTimeToWindowEndSparkNode(SparkExecNode):
    input_node: SparkExecNode
    anchor_time_columns: List[str]
    aggregation_tile_interval_column_map: Dict[str, int]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        output_df = self.input_node.to_dataframe(spark)
        for column in self.anchor_time_columns:
            aggregation_tile_interval = self.aggregation_tile_interval_column_map[column]
            if aggregation_tile_interval != 0:
                output_df = output_df.withColumn(column, F.col(column) + aggregation_tile_interval)
        return output_df


@attrs.frozen
class AddEffectiveTimestampSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_field: str
    effective_timestamp_name: str
    batch_schedule_seconds: int
    is_stream: bool
    data_delay_seconds: int
    is_temporal_aggregate: bool

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)

        # batch_schedule = 0 implies feature table.
        if self.batch_schedule_seconds == 0 or self.is_stream:
            effective_timestamp = F.col(self.timestamp_field)
        else:
            slide_str = f"{self.batch_schedule_seconds} seconds"
            timestamp_col = F.col(self.timestamp_field)
            # Timestamp of temporal aggregate is end of the anchor time window. Subtract 1 micro
            # to get the correct bucket for batch schedule.
            if self.is_temporal_aggregate:
                timestamp_col -= expr("interval 1 microseconds")
            window_spec = F.window(timestamp_col, slide_str, slide_str)
            effective_timestamp = window_spec.end + expr(f"interval {self.data_delay_seconds} seconds")

        df = input_df.withColumn(self.effective_timestamp_name, effective_timestamp)
        return df


@attrs.frozen
class AddDurationSparkNode(SparkExecNode):
    input_node: SparkExecNode
    timestamp_field: str
    duration: pendulum.Duration
    new_column_name: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        return input_df.withColumn(
            self.new_column_name,
            F.col(self.timestamp_field) + expr(f"interval {self.duration.total_seconds()} seconds"),
        )


@attrs.frozen
class SelectDistinctSparkNode(SparkExecNode):
    input_node: SparkExecNode
    columns: List[str]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        return input_df.select(self.columns).distinct()


@attrs.frozen
class AddBooleanPartitionColumnsSparkNode(SparkExecNode):
    input_node: SparkExecNode
    column_to_bool_map: Dict[str, bool]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        for column_name, value in self.column_to_bool_map.items():
            input_df = input_df.withColumn(column_name, F.lit(value))
        return input_df


@attrs.frozen
class ExplodeEventsAndSelectDistinctSparkNode(SparkExecNode):
    """
    Explode the events in the input node and select distinct columns.
    """

    input_node: SparkExecNode
    explode_columns: List[str]
    explode_columns_to_boolean_columns: Dict[str, str]
    timestamp_column: str
    columns_to_ignore: List[str]

    def _add_identifier_columns(self, df: pyspark.sql.DataFrame, explode_column: str) -> pyspark.sql.DataFrame:
        columns_to_add = []
        for column in self.explode_columns:
            identifier = self.explode_columns_to_boolean_columns[column]
            if column == explode_column:
                columns_to_add.append(F.lit(True).alias(identifier))
            else:
                columns_to_add.append(F.lit(False).alias(identifier))
        return df.select("*", *columns_to_add)

    def _select_distinct_for_explode_column(
        self, input_df: pyspark.sql.DataFrame, explode_column: str
    ) -> pyspark.sql.DataFrame:
        new_df = self._add_identifier_columns(input_df, explode_column)
        other_columns = self.columns_to_ignore + [col for col in self.explode_columns if col != explode_column]
        for column in other_columns:
            new_df = new_df.withColumn(column, F.lit(None))
        return new_df.select(*self.columns).distinct()

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)
        union_df = self._select_distinct_for_explode_column(input_df, self.explode_columns[0])
        if len(self.explode_columns) == 1:
            return union_df

        for explode_column in self.explode_columns[1:]:
            intermediate_df = self._select_distinct_for_explode_column(input_df, explode_column)
            union_df = union_df.union(intermediate_df)
        return union_df


@attrs.frozen
class AggregationSecondaryKeyExplodeSparkNode(SparkExecNode):
    """
    This node returns all <entity, aggregation secondary key> pairs from the input node. It's used for retrieving
    secondary key aggregate features without spine.

    This node is similar to AsofSecondaryKeyExplodeSparkNode but doesn't require as-of join. It uses entities and
    anchor time to look back the max aggregation window time to find all secondary key values, and build a spine with
    all of them.
    """

    input_node: SparkExecNode
    join_keys: List[str]
    aggregation_secondary_key: str
    has_lifetime_aggregate: bool
    earliest_anchor_time_from_window_start: Optional[int]

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        input_df = self.input_node.to_dataframe(spark)

        window_start = (
            spark_window.Window.unboundedPreceding
            if self.has_lifetime_aggregate
            else self.earliest_anchor_time_from_window_start
        )

        window_spec = (
            spark_window.Window.partitionBy(self.join_keys)
            .orderBy([anchor_time()])
            .rangeBetween(window_start, spark_window.Window.currentRow)
        )

        # Collect all secondary key values within the aggregation window, and explode them to build a spine.
        # Note the window based `collect_set` applies to every single row, so if a single join key has multiple
        # aggregation secondary keys with the same anchor time, the result dataframe will have multiple same rows for
        # that particular join key and anchor time. We select distinct here to deduplicate.
        input_df = (
            input_df.withColumn(
                TECTON_TEMP_AGGREGATION_SECONDARY_KEY_COL,
                F.collect_set(self.aggregation_secondary_key).over(window_spec),
            )
            .drop(self.aggregation_secondary_key)
            .distinct()
        )
        return input_df.withColumn(
            self.aggregation_secondary_key, F.explode(TECTON_TEMP_AGGREGATION_SECONDARY_KEY_COL)
        ).drop(TECTON_TEMP_AGGREGATION_SECONDARY_KEY_COL)


@attrs.frozen
class AddUniqueIdSparkNode(SparkExecNode):
    """
    Add a '_tecton_unique_id' column to the input node. This column is used to uniquely identify rows in the input node.

    Warning: The generated unique ID is non-deterministic on Spark, so this column may not be safe to join on.
    """

    input_node: SparkExecNode
    column_name: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        return self.input_node.to_dataframe(spark).withColumn(self.column_name, F.monotonically_increasing_id())


def _is_count_feature(feature: feature_view__data_pb2.Aggregate) -> bool:
    return (
        feature.function == AggregationFunction.AGGREGATION_FUNCTION_COUNT
        or feature.function == AggregationFunction.AGGREGATION_FUNCTION_APPROX_COUNT_DISTINCT
    )


@attrs.frozen
class DeriveValidityPeriodSparkNode(SparkExecNode):
    """
    Derives the `valid_from` and `valid_to` columns from a fully aggregated data frame
    and removes duplicates and rows with default values for all aggregation columns.
    """

    input_node: SparkExecNode
    fdw: FeatureDefinitionWrapper
    timestamp_field: str

    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        join_keys = self.fdw.join_keys
        df = self.input_node.to_dataframe(spark)
        features = self.fdw.fv_spec.features
        window = spark_window.Window.partitionBy(*join_keys).orderBy(self.timestamp_field)

        df = df.withColumn(valid_from(), F.col(self.timestamp_field)).withColumn(
            valid_to(), F.lead(self.timestamp_field).over(window)
        )

        feature_value_update_conditions = [
            ~F.col(column).eqNullSafe(F.col(f"_tecton_prev_{column}")) for column in features
        ]
        track_changes_cols = list(features)

        if self.fdw.serving_ttl is not None:
            df = self._expire_ttl(df)

            # if `valid_from` is not equal to the previous `valid_to` (in the case of TTL expiry),
            # we should not merge the rows even though the feature values are equal
            track_changes_cols.append(valid_to())
            feature_value_update_conditions.append(~F.col(valid_from()).eqNullSafe(F.col(f"_tecton_prev_{valid_to()}")))

        previous_value_columns = [
            F.lag(column).over(window).alias(f"_tecton_prev_{column}") for column in track_changes_cols
        ]
        df = df.select("*", *previous_value_columns)

        # Check if any feature column changed in value, set is_new=1 for rows with changes
        df = df.withColumn(
            "_tecton_is_new", F.when(reduce(lambda x, y: x | y, feature_value_update_conditions), 1).otherwise(0)
        )

        # Cumulatively sum the `is_new` column to create "groups".
        # A group is a set of consecutive rows that has the same aggregated value.
        df = df.withColumn("_tecton_group", F.sum("_tecton_is_new").over(window))

        # if valid_to for a group is "null", we've reached the last event and the "max" should be null
        # since we don't know how long the value is valid
        null_check = F.max(F.when(F.col(valid_to()).isNull(), True).otherwise(False))
        df = df.groupBy(*join_keys, *features, "_tecton_group").agg(
            F.min(valid_from()).alias(valid_from()),
            F.when(null_check, F.lit(None)).otherwise(F.max(valid_to())).alias(valid_to()),
        )

        if isinstance(self.fdw.fv_spec, MaterializedFeatureViewSpec):
            if self.fdw.fv_spec.type == MaterializedFeatureViewType.TEMPORAL_AGGREGATE:
                df = self._remove_default_values(df)

        return df.drop("_tecton_group", self.timestamp_field)

    def _remove_default_values(self, df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
        """
        Removes rows with default values for all aggregation columns.
        """
        aggregate_features = self.fdw.fv_spec.aggregate_features
        is_default_value_conditions = []
        if self.fdw.aggregation_secondary_key:
            # For secondary key aggs, we remove rows that have empty rollup output lists
            column_names = [rollup_output.name for rollup_output in self.fdw.fv_spec.secondary_key_rollup_outputs]
            is_default_value_conditions.extend([F.size(F.col(column)) == 0 for column in column_names])
        else:
            for feature in aggregate_features:
                column_name = feature.output_feature_name
                if isinstance(df.schema[column_name].dataType, ArrayType):
                    all_counts_empty = F.size(F.array_except(column_name, F.array(F.lit(0)))) == 0
                    all_nulls = F.size(F.array_except(column_name, F.array(F.lit(None)))) == 0
                    if _is_count_feature(feature):
                        condition = all_counts_empty | all_nulls
                    else:
                        condition = all_nulls
                else:
                    if _is_count_feature(feature):
                        condition = F.col(column_name).isNull() | (F.col(column_name) == 0)
                    else:
                        condition = F.col(column_name).isNull()
                is_default_value_conditions.append(condition)

        return df.filter(~reduce(lambda x, y: x & y, is_default_value_conditions))

    def _expire_ttl(self, df: pyspark.sql.DataFrame) -> pyspark.sql.DataFrame:
        """
        Trims `valid_to` values according to TTL for non-aggregate feature views
        """
        ttl_seconds = self.fdw.serving_ttl.total_seconds()
        should_trim_valid_to = F.col(valid_to()).isNull() | (
            (F.col(valid_to()).cast(LongType()) - F.col(valid_from()).cast(LongType())) > ttl_seconds
        )
        df = df.withColumn(
            valid_to(),
            F.when(should_trim_valid_to, F.col(valid_from()) + expr(f"INTERVAL {ttl_seconds} SECONDS")).otherwise(
                F.col(valid_to())
            ),
        )

        return df



from tecton_proto.auditlog import metadata__client_pb2 as _metadata__client_pb2
from tecton_proto.auth import service__client_pb2 as _service__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CreateSecretScopeRequest(_message.Message):
    __slots__ = ["scope"]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    scope: str
    def __init__(self, scope: _Optional[str] = ...) -> None: ...

class CreateSecretScopeResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class DeleteSecretRequest(_message.Message):
    __slots__ = ["key", "scope"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    key: str
    scope: str
    def __init__(self, scope: _Optional[str] = ..., key: _Optional[str] = ...) -> None: ...

class DeleteSecretResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class DeleteSecretScopeRequest(_message.Message):
    __slots__ = ["scope"]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    scope: str
    def __init__(self, scope: _Optional[str] = ...) -> None: ...

class DeleteSecretScopeResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class GetSecretValueRequest(_message.Message):
    __slots__ = ["key", "scope"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    key: str
    scope: str
    def __init__(self, scope: _Optional[str] = ..., key: _Optional[str] = ...) -> None: ...

class GetSecretValueResponse(_message.Message):
    __slots__ = ["value"]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: str
    def __init__(self, value: _Optional[str] = ...) -> None: ...

class ListSecretScopesRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ListSecretScopesResponse(_message.Message):
    __slots__ = ["scopes"]
    SCOPES_FIELD_NUMBER: _ClassVar[int]
    scopes: _containers.RepeatedCompositeFieldContainer[SecretScope]
    def __init__(self, scopes: _Optional[_Iterable[_Union[SecretScope, _Mapping]]] = ...) -> None: ...

class ListSecretsRequest(_message.Message):
    __slots__ = ["scope"]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    scope: str
    def __init__(self, scope: _Optional[str] = ...) -> None: ...

class ListSecretsResponse(_message.Message):
    __slots__ = ["keys"]
    KEYS_FIELD_NUMBER: _ClassVar[int]
    keys: _containers.RepeatedCompositeFieldContainer[SecretKey]
    def __init__(self, keys: _Optional[_Iterable[_Union[SecretKey, _Mapping]]] = ...) -> None: ...

class PutSecretValueRequest(_message.Message):
    __slots__ = ["key", "scope", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    SCOPE_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    scope: str
    value: str
    def __init__(self, scope: _Optional[str] = ..., key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class PutSecretValueResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class SecretKey(_message.Message):
    __slots__ = ["name"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class SecretScope(_message.Message):
    __slots__ = ["name"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...



from enum import Enum
from typing import Iterable
from typing import Optional

import attrs

from tecton_core.embeddings.model_artifacts import ModelArtifactProvider
from tecton_core.offline_store import OfflineStoreOptionsProvider
from tecton_core.secret_management import SecretResolver


class QueryTreeStep(Enum):
    """Query trees are composed of steps.

    Each step may have its own compute, and may stage its results in preparation for the next step.

    Query trees do not have to have all steps. For example, a materialization query tree for a batch feature view will
    not have any nodes in the ODFV step.
    """

    # Runs data source scans.
    DATA_SOURCE = 1
    # Runs feature view transformations to produce un-aggregated feature data.
    PIPELINE = 2
    # Runs model inference (only used by text embeddings right now).
    MODEL_INFERENCE = 3
    # Runs partial aggregations, full aggregations, and the as-of join.
    AGGREGATION = 4
    # Runs on-demand transformations.
    ODFV = 5


@attrs.frozen
class ExecutionContext:
    offline_store_options_providers: Iterable[OfflineStoreOptionsProvider]
    secret_resolver: Optional[SecretResolver] = None
    model_artifact_provider: Optional[ModelArtifactProvider] = None



import re
from typing import List

import attrs

from tecton_core import feature_set_config
from tecton_core.errors import TectonInternalError
from tecton_core.errors import TectonValidationError
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.realtime_context import REQUEST_TIMESTAMP_FIELD_NAME
from tecton_core.schema import Schema
from tecton_proto.args import pipeline__client_pb2 as pipeline_pb2
from tecton_proto.args.pipeline__client_pb2 import PipelineNode


_COLUMN_REFERENCE_NAMESPACE_SEPARATOR = re.compile(r"__|\.")


@attrs.define
class ColumnReferenceResolver:
    """
    A class to resolve column reference strings in features to their internal column name strings in the query tree.
    """

    events_df_timestamp_field: str
    input_node_column_names: List[str]

    def _input_node_has_column(self, column_name: str) -> bool:
        return column_name in self.input_node_column_names

    @staticmethod
    def _get_feature_view_node_internal_column_name(
        source_fv_pipeline_node: PipelineNode, feature_name: str, fdw: FeatureDefinitionWrapper
    ) -> str:
        # the source fv is in the fco_container, and it has no dependents, so there will only be 1 item in the result
        dependent_fv_dac = feature_set_config.find_dependent_feature_set_items(
            fdw.fco_container, source_fv_pipeline_node, {}, fdw.id
        )[0]
        if feature_name not in dependent_fv_dac.features:
            msg = f"Feature {feature_name} not found in source feature view {dependent_fv_dac.name}."
            raise TectonInternalError(msg)
        return f"{dependent_fv_dac.namespace}{dependent_fv_dac.feature_definition.namespace_separator}{feature_name}"

    def _get_realtime_context_internal_column_name(self, feature_name: str) -> str:
        if feature_name == REQUEST_TIMESTAMP_FIELD_NAME:
            if not self._input_node_has_column(self.events_df_timestamp_field):
                msg = f"Timestamp field {self.events_df_timestamp_field} not found in input events df. Missing for use in realtime context request_timestamp."
                raise TectonValidationError(msg)
            return self.events_df_timestamp_field
        msg = f"Context attribute {feature_name} not available in realtime context."
        raise TectonInternalError(msg)

    @staticmethod
    def _get_request_data_source_internal_column_name(pipeline_node: PipelineNode, feature_name: str) -> str:
        request_context_schema = pipeline_node.request_data_source_node.request_context.tecton_schema
        schema = Schema(request_context_schema)
        if feature_name in schema.column_names():
            return feature_name
        msg = f"Feature name {feature_name} not found in request data source {pipeline_node.request_data_source_node.input_name}"
        raise TectonValidationError(msg)

    @staticmethod
    def _get_supported_pipeline_node_name(pipeline_node: PipelineNode) -> str:
        if pipeline_node.HasField("feature_view_node"):
            return pipeline_node.feature_view_node.input_name
        elif pipeline_node.HasField("context_node"):
            return pipeline_node.context_node.input_name
        elif pipeline_node.HasField("request_data_source_node"):
            return pipeline_node.request_data_source_node.input_name
        else:
            node_type = pipeline_node.WhichOneof("node_type")
            msg = f"Pipeline node type {node_type} not supported."
            raise TectonInternalError(msg)

    @staticmethod
    def _split_column_reference(column_reference: str) -> List[str]:
        splits = _COLUMN_REFERENCE_NAMESPACE_SEPARATOR.split(column_reference)
        if len(splits) != 2:
            msg = f"Expected column_reference to contain exactly one '__' or '.', but got '{column_reference}' instead."
            raise ValueError(msg)
        return splits

    def get_internal_column_name(self, column_reference: str, fdw: FeatureDefinitionWrapper) -> str:
        """Resolves a column reference string in an attribute or calculation to its corresponding internal column name.

        Attribute and Calculation features can reference columns from their sources like `my_batch_fv__transaction_amt`,
        or `my_request_source__client_app_version`. In the intermediate nodes of the query tree, these data
        columns have internal names like `_udf_internal_my_batch_fv_ab353c4597780f6c7ff33f1a3655a4ca__transaction_amt`.

        This method takes an input column reference, validates it by confirming the column reference refers to a column
        in the feature's sources, and then resolves the column reference string to the internal column name,
        so it can be used in the composed duck db or spark sql query.

        Examples:
             - `my_batch_fv__transaction_amt` -> `_udf_internal_my_batch_fv_ab353c4597780f6c7ff33f1a3655a4ca__transaction_amt`
             - `context__request_timestamp` -> `TIMESTAMP` (timestamp_key in the events df)
        """
        source_name, feature_name = self._split_column_reference(column_reference)
        pipeline_node = fdw.pipeline.root
        if not pipeline_node.HasField("join_inputs_node"):
            msg = "Pipeline root node is not join_inputs_node type. Cannot resolve internal columns from other node types."
            raise TectonInternalError(msg)
        join_inputs_node = pipeline_node.join_inputs_node

        source_pipeline_nodes = [
            input_node
            for input_node in join_inputs_node.nodes
            if self._get_supported_pipeline_node_name(input_node) == source_name
        ]
        if len(source_pipeline_nodes) != 1:
            msg = f"Found {len(source_pipeline_nodes)} input pipeline nodes for the source {source_name} for column reference {column_reference}. Source must match 1 input pipeline node."
            raise TectonInternalError(msg)
        source_pipeline_node = source_pipeline_nodes[0]

        internal_column_name = None
        if source_pipeline_node.HasField("feature_view_node"):
            internal_column_name = self._get_feature_view_node_internal_column_name(
                source_pipeline_node, feature_name, fdw
            )
        elif (
            source_pipeline_node.HasField("context_node")
            and source_pipeline_node.context_node.context_type == pipeline_pb2.CONTEXT_TYPE_REALTIME
        ):
            internal_column_name = self._get_realtime_context_internal_column_name(feature_name)
        elif source_pipeline_node.HasField("request_data_source_node"):
            internal_column_name = self._get_request_data_source_internal_column_name(
                source_pipeline_node, feature_name
            )
        else:
            node_type = source_pipeline_node.WhichOneof("node_type")
            msg = f"Failed to extract Features. Pipeline node type {node_type} not supported."
            raise TectonInternalError(msg)

        # check if the internal_column is in the input columns
        # Could extend this to check membership in input query or input dataframe for spark and sql, respectively
        if not internal_column_name or not self._input_node_has_column(internal_column_name):
            msg = f"Internal column {internal_column_name} for column_reference {column_reference} not found in input to Feature Extraction."
            raise TectonInternalError(msg)

        return internal_column_name



from datetime import datetime
from typing import Tuple

import boto3

import tecton_core.tecton_pendulum as pendulum
from tecton_core import conf
from tecton_core import query_consts
from tecton_core.compute_mode import ComputeMode
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import FeatureTimeFilterNode
from tecton_core.query.nodes import OfflineStoreScanNode
from tecton_core.query.nodes import StagingNode
from tecton_core.query.nodes import TakeLastRowNode
from tecton_materialization.common.task_params import TimeInterval


# TODO(liangqi): Use this function in the Spark code to unify the behaviors.
# https://tecton.atlassian.net/browse/BC-3701
def get_bootstrap_bulk_backfill_qt(
    fd: FeatureDefinitionWrapper, start_time: datetime, end_time: datetime
) -> Tuple[NodeRef, TimeInterval]:
    conf.set("CLUSTER_REGION", boto3.Session().region_name)

    tree = OfflineStoreScanNode(
        dialect=Dialect.DUCKDB,
        compute_mode=ComputeMode.RIFT,
        feature_definition_wrapper=fd,
        partition_time_filter=pendulum.Period(start_time, end_time),
    ).as_ref()

    tree = StagingNode(
        dialect=Dialect.ARROW,
        compute_mode=ComputeMode.RIFT,
        input_node=tree,
        staging_table_name=f"offline_store_{fd.name}",
    ).as_ref()

    tree = FeatureTimeFilterNode(
        dialect=Dialect.DUCKDB,
        compute_mode=ComputeMode.RIFT,
        input_node=tree,
        feature_data_time_limits=pendulum.Period(start_time, end_time),
        policy=fd.time_range_policy,
        start_timestamp_field=fd.timestamp_key if fd.is_temporal else query_consts.anchor_time(),
        end_timestamp_field=fd.timestamp_key if fd.is_temporal else query_consts.anchor_time(),
        is_timestamp_format=fd.is_temporal,
        feature_store_format_version=fd.get_feature_store_format_version,
    ).as_ref()

    if fd.is_temporal:
        tree = TakeLastRowNode(
            dialect=Dialect.DUCKDB,
            compute_mode=ComputeMode.RIFT,
            input_node=tree,
            partition_by_columns=fd.join_keys,
            order_by_column=fd.timestamp_key,
        ).as_ref()

    return tree



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/server_groups/server_group_states__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import container_image__client_pb2 as tecton__proto_dot_common_dot_container__image____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.common import server_group_status__client_pb2 as tecton__proto_dot_common_dot_server__group__status____client__pb2
from tecton_proto.common import server_group_type__client_pb2 as tecton__proto_dot_common_dot_server__group__type____client__pb2
from tecton_proto.realtime import instance_group__client_pb2 as tecton__proto_dot_realtime_dot_instance__group____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n<tecton_proto/server_groups/server_group_states__client.proto\x12\x1atecton_proto.server_groups\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x31tecton_proto/common/container_image__client.proto\x1a$tecton_proto/common/id__client.proto\x1a\x35tecton_proto/common/server_group_status__client.proto\x1a\x33tecton_proto/common/server_group_type__client.proto\x1a\x32tecton_proto/realtime/instance_group__client.proto\"\xfe\x07\n\x10ServerGroupState\x12J\n\x15server_group_state_id\x18\x0e \x01(\x0b\x32\x17.tecton_proto.common.IdR\x12serverGroupStateId\x12?\n\x0fserver_group_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rserverGroupId\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12>\n\x06status\x18\x03 \x01(\x0e\x32&.tecton_proto.common.ServerGroupStatusR\x06status\x12\x38\n\x04type\x18\x04 \x01(\x0e\x32$.tecton_proto.common.ServerGroupTypeR\x04type\x12\\\n\x12\x61utoscaling_policy\x18\x11 \x01(\x0b\x32-.tecton_proto.server_groups.AutoscalingPolicyR\x11\x61utoscalingPolicy\x12%\n\x0estatus_details\x18\x05 \x01(\tR\rstatusDetails\x12\x39\n\ncreated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x42\n\x0flast_updated_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\rlastUpdatedAt\x12\x1b\n\tmin_nodes\x18\x0f \x01(\x05R\x08minNodes\x12\x1b\n\tmax_nodes\x18\x10 \x01(\x05R\x08maxNodes\x12#\n\rdesired_nodes\x18\r \x01(\x05R\x0c\x64\x65siredNodes\x12x\n\x1ctransform_server_group_state\x18\t \x01(\x0b\x32\x35.tecton_proto.server_groups.TransformServerGroupStateH\x00R\x19transformServerGroupState\x12r\n\x1a\x66\x65\x61ture_server_group_state\x18\n \x01(\x0b\x32\x33.tecton_proto.server_groups.FeatureServerGroupStateH\x00R\x17\x66\x65\x61tureServerGroupState\x12\x1c\n\tworkspace\x18\x0b \x01(\tR\tworkspace\x12\x45\n\x12workspace_state_id\x18\x0c \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10workspaceStateIdB\x13\n\x11server_group_typeJ\x04\x08\x12\x10\x13\"\xd4\x01\n\x11\x41utoscalingPolicy\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12/\n\x13\x61utoscaling_enabled\x18\x02 \x01(\x08R\x12\x61utoscalingEnabled\x12h\n\x16target_tracking_policy\x18\x04 \x01(\x0b\x32\x30.tecton_proto.server_groups.TargetTrackingPolicyH\x00R\x14targetTrackingPolicyB\x10\n\x0escaling_policy\"\x9d\x02\n\x14TargetTrackingPolicy\x12!\n\x0ctarget_value\x18\x01 \x01(\x01R\x0btargetValue\x12*\n\x11scale_in_disabled\x18\x03 \x01(\x08R\x0fscaleInDisabled\x12[\n\x11predefined_metric\x18\x04 \x01(\x0b\x32,.tecton_proto.server_groups.PredefinedMetricH\x00R\x10predefinedMetric\x12O\n\rcustom_metric\x18\x05 \x01(\x0b\x32(.tecton_proto.server_groups.CustomMetricH\x00R\x0c\x63ustomMetricB\x08\n\x06metric\"&\n\x10PredefinedMetric\x12\x12\n\x04type\x18\x01 \x01(\tR\x04type\"T\n\x0c\x43ustomMetric\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x1c\n\tstatistic\x18\x02 \x01(\tR\tstatistic\x12\x12\n\x04unit\x18\x03 \x01(\tR\x04unit\"\xa6\x05\n\x19TransformServerGroupState\x12%\n\x0e\x65nvironment_id\x18\x01 \x01(\tR\renvironmentId\x12)\n\x10\x65nvironment_name\x18\x02 \x01(\tR\x0f\x65nvironmentName\x12\x42\n\nimage_info\x18\x03 \x01(\x0b\x32#.tecton_proto.common.ContainerImageR\timageInfo\x12\x84\x01\n\x15\x65nvironment_variables\x18\x04 \x03(\x0b\x32O.tecton_proto.server_groups.TransformServerGroupState.EnvironmentVariablesEntryR\x14\x65nvironmentVariables\x12\\\n\x12\x61utoscaling_policy\x18\x05 \x01(\x0b\x32-.tecton_proto.server_groups.AutoscalingPolicyR\x11\x61utoscalingPolicy\x12^\n\x15instance_group_handle\x18\x06 \x01(\x0b\x32*.tecton_proto.realtime.InstanceGroupHandleR\x13instanceGroupHandle\x12_\n\x0esecrets_config\x18\x08 \x01(\x0b\x32\x38.tecton_proto.server_groups.TransformServerSecretsConfigR\rsecretsConfig\x1aG\n\x19\x45nvironmentVariablesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01J\x04\x08\x07\x10\x08\"\xd7\x01\n\x1cTransformServerSecretsConfig\x12,\n\x12service_account_id\x18\x04 \x01(\tR\x10serviceAccountId\x12\x44\n\x1fservice_account_key_secret_name\x18\x02 \x01(\tR\x1bserviceAccountKeySecretName\x12=\n\x0clast_updated\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0blastUpdatedJ\x04\x08\x01\x10\x02\"\xfb\x02\n\x1e\x46\x65\x61tureServerGroupAWSResources\x12\x43\n\x1eloadbalancer_listener_rule_arn\x18\x01 \x01(\tR\x1bloadbalancerListenerRuleArn\x12\x30\n\x14\x61utoscaling_group_id\x18\x02 \x01(\tR\x12\x61utoscalingGroupId\x12(\n\x10target_group_arn\x18\x03 \x01(\tR\x0etargetGroupArn\x12,\n\x12launch_template_id\x18\x04 \x01(\tR\x10launchTemplateId\x12\x43\n\x1elaunch_template_active_version\x18\x05 \x01(\tR\x1blaunchTemplateActiveVersion\x12\x45\n\x1flaunch_template_pending_version\x18\x06 \x01(\tR\x1claunchTemplatePendingVersion\"\xaf\x04\n\x18\x46\x65\x61tureServerGroupConfig\x12#\n\rinstance_type\x18\x01 \x01(\tR\x0cinstanceType\x12(\n\x10\x64ocker_image_fqn\x18\x02 \x01(\tR\x0e\x64ockerImageFqn\x12\x1b\n\thttp_port\x18\x03 \x01(\x05R\x08httpPort\x12\x33\n\x16http_health_check_path\x18\x04 \x01(\tR\x13httpHealthCheckPath\x12!\n\x0cmetrics_port\x18\x05 \x01(\x05R\x0bmetricsPort\x12/\n\x13\x61utoscaling_enabled\x18\x06 \x01(\x08R\x12\x61utoscalingEnabled\x12#\n\rmin_instances\x18\x07 \x01(\x05R\x0cminInstances\x12#\n\rmax_instances\x18\x08 \x01(\x05R\x0cmaxInstances\x12+\n\x11\x64\x65sired_instances\x18\t \x01(\x05R\x10\x64\x65siredInstances\x12g\n\x0b\x65nvironment\x18\n \x03(\x0b\x32\x45.tecton_proto.server_groups.FeatureServerGroupConfig.EnvironmentEntryR\x0b\x65nvironment\x1a>\n\x10\x45nvironmentEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xc7\x02\n\x17\x46\x65\x61tureServerGroupState\x12Y\n\ractive_config\x18\x01 \x01(\x0b\x32\x34.tecton_proto.server_groups.FeatureServerGroupConfigR\x0c\x61\x63tiveConfig\x12[\n\x0epending_config\x18\x02 \x01(\x0b\x32\x34.tecton_proto.server_groups.FeatureServerGroupConfigR\rpendingConfig\x12\x61\n\raws_resources\x18\x03 \x01(\x0b\x32:.tecton_proto.server_groups.FeatureServerGroupAWSResourcesH\x00R\x0c\x61wsResourcesB\x11\n\x0f\x63loud_resources*a\n\x11ScalingPolicyType\x12#\n\x1fSCALING_POLICY_TYPE_UNSPECIFIED\x10\x00\x12\'\n#SCALING_POLICY_TYPE_TARGET_TRACKING\x10\x01\x42\x1b\n\x17\x63om.tecton.servergroupsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.server_groups.server_group_states__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\027com.tecton.servergroupsP\001'
  _TRANSFORMSERVERGROUPSTATE_ENVIRONMENTVARIABLESENTRY._options = None
  _TRANSFORMSERVERGROUPSTATE_ENVIRONMENTVARIABLESENTRY._serialized_options = b'8\001'
  _FEATURESERVERGROUPCONFIG_ENVIRONMENTENTRY._options = None
  _FEATURESERVERGROUPCONFIG_ENVIRONMENTENTRY._serialized_options = b'8\001'
  _SCALINGPOLICYTYPE._serialized_start=4201
  _SCALINGPOLICYTYPE._serialized_end=4298
  _SERVERGROUPSTATE._serialized_start=375
  _SERVERGROUPSTATE._serialized_end=1397
  _AUTOSCALINGPOLICY._serialized_start=1400
  _AUTOSCALINGPOLICY._serialized_end=1612
  _TARGETTRACKINGPOLICY._serialized_start=1615
  _TARGETTRACKINGPOLICY._serialized_end=1900
  _PREDEFINEDMETRIC._serialized_start=1902
  _PREDEFINEDMETRIC._serialized_end=1940
  _CUSTOMMETRIC._serialized_start=1942
  _CUSTOMMETRIC._serialized_end=2026
  _TRANSFORMSERVERGROUPSTATE._serialized_start=2029
  _TRANSFORMSERVERGROUPSTATE._serialized_end=2707
  _TRANSFORMSERVERGROUPSTATE_ENVIRONMENTVARIABLESENTRY._serialized_start=2630
  _TRANSFORMSERVERGROUPSTATE_ENVIRONMENTVARIABLESENTRY._serialized_end=2701
  _TRANSFORMSERVERSECRETSCONFIG._serialized_start=2710
  _TRANSFORMSERVERSECRETSCONFIG._serialized_end=2925
  _FEATURESERVERGROUPAWSRESOURCES._serialized_start=2928
  _FEATURESERVERGROUPAWSRESOURCES._serialized_end=3307
  _FEATURESERVERGROUPCONFIG._serialized_start=3310
  _FEATURESERVERGROUPCONFIG._serialized_end=3869
  _FEATURESERVERGROUPCONFIG_ENVIRONMENTENTRY._serialized_start=3807
  _FEATURESERVERGROUPCONFIG_ENVIRONMENTENTRY._serialized_end=3869
  _FEATURESERVERGROUPSTATE._serialized_start=3872
  _FEATURESERVERGROUPSTATE._serialized_end=4199
# @@protoc_insertion_point(module_scope)



from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

CUSTOM_COMPARATOR_AGGREGATION_NAME: CustomComparator
CUSTOM_COMPARATOR_BATCH_SCHEDULE: CustomComparator
CUSTOM_COMPARATOR_DIFF_OVERRIDE_JOIN_KEYS_AS_MAP: CustomComparator
CUSTOM_COMPARATOR_DISPLAY_NOTSET: CustomComparator
CUSTOM_COMPARATOR_ENTITY_JOIN_KEYS: CustomComparator
CUSTOM_COMPARATOR_FEATURE_PARAM_UPGRADE: CustomComparator
CUSTOM_COMPARATOR_FILTERED_SOURCE: CustomComparator
CUSTOM_COMPARATOR_OFFLINE_STORE: CustomComparator
CUSTOM_COMPARATOR_OFFLINE_STORE_LEGACY: CustomComparator
CUSTOM_COMPARATOR_ONLINE_OFFLINE_ENABLED: CustomComparator
CUSTOM_COMPARATOR_OPTION_VALUE_WITH_REDACTION: CustomComparator
CUSTOM_COMPARATOR_REQUEST_SOURCE: CustomComparator
CUSTOM_COMPARATOR_STREAM_PROCESSING_MODE: CustomComparator
CUSTOM_COMPARATOR_TIMESTAMP_FIELD: CustomComparator
CUSTOM_COMPARATOR_TIME_WINDOW: CustomComparator
CUSTOM_COMPARATOR_TIME_WINDOW_LEGACY: CustomComparator
CUSTOM_COMPARATOR_UNITY_CATALOG_ACCESS_MODE: CustomComparator
CUSTOM_COMPARATOR_UNSET: CustomComparator
DESCRIPTOR: _descriptor.FileDescriptor
DIFF_OPTIONS_FIELD_NUMBER: _ClassVar[int]
FCO_PROPERTY_RENDERING_TYPE_HIDDEN: FcoPropertyRenderingType
FCO_PROPERTY_RENDERING_TYPE_ONLY_DECLARED: FcoPropertyRenderingType
FCO_PROPERTY_RENDERING_TYPE_PLAIN_TEXT: FcoPropertyRenderingType
FCO_PROPERTY_RENDERING_TYPE_PYTHON: FcoPropertyRenderingType
FCO_PROPERTY_RENDERING_TYPE_REDACTED: FcoPropertyRenderingType
FCO_PROPERTY_RENDERING_TYPE_SQL: FcoPropertyRenderingType
FCO_PROPERTY_RENDERING_TYPE_UNSPECIFIED: FcoPropertyRenderingType
INPLACE: UpdateStrategy
INPLACE_ON_ADD: UpdateStrategy
INPLACE_ON_REMOVE: UpdateStrategy
ONE_WAY_INPLACE_ON_ADD: UpdateStrategy
PASSIVE: UpdateStrategy
RECREATE: UpdateStrategy
RECREATE_UNLESS_SUPPRESSED: UpdateStrategy
RECREATE_UNLESS_SUPPRESSED_INVALIDATE_CHECKPOINTS: UpdateStrategy
RECREATE_UNLESS_SUPPRESSED_RESTART_STREAM: UpdateStrategy
diff_options: _descriptor.FieldDescriptor

class DiffOptions(_message.Message):
    __slots__ = ["custom_comparator", "hide_path", "rename", "rendering_type", "update"]
    CUSTOM_COMPARATOR_FIELD_NUMBER: _ClassVar[int]
    HIDE_PATH_FIELD_NUMBER: _ClassVar[int]
    RENAME_FIELD_NUMBER: _ClassVar[int]
    RENDERING_TYPE_FIELD_NUMBER: _ClassVar[int]
    UPDATE_FIELD_NUMBER: _ClassVar[int]
    custom_comparator: CustomComparator
    hide_path: bool
    rename: FieldRenameConfig
    rendering_type: FcoPropertyRenderingType
    update: UpdateStrategy
    def __init__(self, update: _Optional[_Union[UpdateStrategy, str]] = ..., hide_path: bool = ..., rendering_type: _Optional[_Union[FcoPropertyRenderingType, str]] = ..., custom_comparator: _Optional[_Union[CustomComparator, str]] = ..., rename: _Optional[_Union[FieldRenameConfig, _Mapping]] = ...) -> None: ...

class FieldRenameConfig(_message.Message):
    __slots__ = ["cutover_version", "former_name"]
    CUTOVER_VERSION_FIELD_NUMBER: _ClassVar[int]
    FORMER_NAME_FIELD_NUMBER: _ClassVar[int]
    cutover_version: str
    former_name: str
    def __init__(self, former_name: _Optional[str] = ..., cutover_version: _Optional[str] = ...) -> None: ...

class UpdateStrategy(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class FcoPropertyRenderingType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class CustomComparator(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import timestamp_pb2 as _timestamp_pb2
from tecton_proto.auditlog import metadata__client_pb2 as _metadata__client_pb2
from tecton_proto.auth import principal__client_pb2 as _principal__client_pb2
from tecton_proto.auth import resource_role_assignments__client_pb2 as _resource_role_assignments__client_pb2
from tecton_proto.auth import service__client_pb2 as _service__client_pb2
from tecton_proto.data import principal_group__client_pb2 as _principal_group__client_pb2
from tecton_proto.data import user__client_pb2 as _user__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AddPrincipalGroupMember(_message.Message):
    __slots__ = ["principal_id", "principal_type"]
    PRINCIPAL_ID_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_TYPE_FIELD_NUMBER: _ClassVar[int]
    principal_id: str
    principal_type: _principal__client_pb2.PrincipalType
    def __init__(self, principal_type: _Optional[_Union[_principal__client_pb2.PrincipalType, str]] = ..., principal_id: _Optional[str] = ...) -> None: ...

class AddPrincipalGroupMembersRequest(_message.Message):
    __slots__ = ["members", "principal_group_id"]
    MEMBERS_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    members: _containers.RepeatedCompositeFieldContainer[AddPrincipalGroupMember]
    principal_group_id: str
    def __init__(self, principal_group_id: _Optional[str] = ..., members: _Optional[_Iterable[_Union[AddPrincipalGroupMember, _Mapping]]] = ...) -> None: ...

class AddPrincipalGroupMembersResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class CreatePrincipalGroupRequest(_message.Message):
    __slots__ = ["description", "idp_mapping_names", "name"]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    IDP_MAPPING_NAMES_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    description: str
    idp_mapping_names: _containers.RepeatedScalarFieldContainer[str]
    name: str
    def __init__(self, name: _Optional[str] = ..., description: _Optional[str] = ..., idp_mapping_names: _Optional[_Iterable[str]] = ...) -> None: ...

class CreatePrincipalGroupResponse(_message.Message):
    __slots__ = ["principal_group"]
    PRINCIPAL_GROUP_FIELD_NUMBER: _ClassVar[int]
    principal_group: _principal_group__client_pb2.PrincipalGroup
    def __init__(self, principal_group: _Optional[_Union[_principal_group__client_pb2.PrincipalGroup, _Mapping]] = ...) -> None: ...

class DeletePrincipalGroupsRequest(_message.Message):
    __slots__ = ["ids"]
    IDS_FIELD_NUMBER: _ClassVar[int]
    ids: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, ids: _Optional[_Iterable[str]] = ...) -> None: ...

class DeletePrincipalGroupsResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ListPrincipalGroupMembersRequest(_message.Message):
    __slots__ = ["id", "principal_types", "search"]
    ID_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_TYPES_FIELD_NUMBER: _ClassVar[int]
    SEARCH_FIELD_NUMBER: _ClassVar[int]
    id: str
    principal_types: _containers.RepeatedScalarFieldContainer[_principal__client_pb2.PrincipalType]
    search: str
    def __init__(self, id: _Optional[str] = ..., search: _Optional[str] = ..., principal_types: _Optional[_Iterable[_Union[_principal__client_pb2.PrincipalType, str]]] = ...) -> None: ...

class ListPrincipalGroupMembersResponse(_message.Message):
    __slots__ = ["members"]
    MEMBERS_FIELD_NUMBER: _ClassVar[int]
    members: _containers.RepeatedCompositeFieldContainer[PrincipalGroupMember]
    def __init__(self, members: _Optional[_Iterable[_Union[PrincipalGroupMember, _Mapping]]] = ...) -> None: ...

class ListPrincipalGroupsDetailedRequest(_message.Message):
    __slots__ = ["ids", "search"]
    IDS_FIELD_NUMBER: _ClassVar[int]
    SEARCH_FIELD_NUMBER: _ClassVar[int]
    ids: _containers.RepeatedScalarFieldContainer[str]
    search: str
    def __init__(self, search: _Optional[str] = ..., ids: _Optional[_Iterable[str]] = ...) -> None: ...

class ListPrincipalGroupsDetailedResponse(_message.Message):
    __slots__ = ["groups"]
    GROUPS_FIELD_NUMBER: _ClassVar[int]
    groups: _containers.RepeatedCompositeFieldContainer[PrincipalGroupDetailed]
    def __init__(self, groups: _Optional[_Iterable[_Union[PrincipalGroupDetailed, _Mapping]]] = ...) -> None: ...

class ListPrincipalGroupsForPrincipalRequest(_message.Message):
    __slots__ = ["principal_id", "principal_type", "search"]
    PRINCIPAL_ID_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_TYPE_FIELD_NUMBER: _ClassVar[int]
    SEARCH_FIELD_NUMBER: _ClassVar[int]
    principal_id: str
    principal_type: _principal__client_pb2.PrincipalType
    search: str
    def __init__(self, principal_type: _Optional[_Union[_principal__client_pb2.PrincipalType, str]] = ..., principal_id: _Optional[str] = ..., search: _Optional[str] = ...) -> None: ...

class ListPrincipalGroupsForPrincipalResponse(_message.Message):
    __slots__ = ["groups"]
    GROUPS_FIELD_NUMBER: _ClassVar[int]
    groups: _containers.RepeatedCompositeFieldContainer[PrincipalGroupForPrincipal]
    def __init__(self, groups: _Optional[_Iterable[_Union[PrincipalGroupForPrincipal, _Mapping]]] = ...) -> None: ...

class ListUsersRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class ListUsersResponse(_message.Message):
    __slots__ = ["users"]
    USERS_FIELD_NUMBER: _ClassVar[int]
    users: _containers.RepeatedCompositeFieldContainer[_user__client_pb2.User]
    def __init__(self, users: _Optional[_Iterable[_Union[_user__client_pb2.User, _Mapping]]] = ...) -> None: ...

class PrincipalGroupDetailed(_message.Message):
    __slots__ = ["account_type", "created_at", "created_by", "description", "id", "idp_mapping_names", "is_membership_editable", "name", "num_members", "workspaces"]
    ACCOUNT_TYPE_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    IDP_MAPPING_NAMES_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    IS_MEMBERSHIP_EDITABLE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    NUM_MEMBERS_FIELD_NUMBER: _ClassVar[int]
    WORKSPACES_FIELD_NUMBER: _ClassVar[int]
    account_type: str
    created_at: _timestamp_pb2.Timestamp
    created_by: _principal__client_pb2.PrincipalBasic
    description: str
    id: str
    idp_mapping_names: _containers.RepeatedScalarFieldContainer[str]
    is_membership_editable: bool
    name: str
    num_members: int
    workspaces: _containers.RepeatedCompositeFieldContainer[_resource_role_assignments__client_pb2.ResourceAndRoleAssignmentsV2]
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., idp_mapping_names: _Optional[_Iterable[str]] = ..., is_membership_editable: bool = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., account_type: _Optional[str] = ..., num_members: _Optional[int] = ..., workspaces: _Optional[_Iterable[_Union[_resource_role_assignments__client_pb2.ResourceAndRoleAssignmentsV2, _Mapping]]] = ..., created_by: _Optional[_Union[_principal__client_pb2.PrincipalBasic, _Mapping]] = ...) -> None: ...

class PrincipalGroupForPrincipal(_message.Message):
    __slots__ = ["group", "is_from_idp_mapping"]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    IS_FROM_IDP_MAPPING_FIELD_NUMBER: _ClassVar[int]
    group: _principal_group__client_pb2.PrincipalGroup
    is_from_idp_mapping: bool
    def __init__(self, group: _Optional[_Union[_principal_group__client_pb2.PrincipalGroup, _Mapping]] = ..., is_from_idp_mapping: bool = ...) -> None: ...

class PrincipalGroupMember(_message.Message):
    __slots__ = ["is_from_idp_mapping", "principal"]
    IS_FROM_IDP_MAPPING_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_FIELD_NUMBER: _ClassVar[int]
    is_from_idp_mapping: bool
    principal: _principal__client_pb2.PrincipalBasic
    def __init__(self, principal: _Optional[_Union[_principal__client_pb2.PrincipalBasic, _Mapping]] = ..., is_from_idp_mapping: bool = ...) -> None: ...

class RemovePrincipalGroupMember(_message.Message):
    __slots__ = ["principal_id", "principal_type"]
    PRINCIPAL_ID_FIELD_NUMBER: _ClassVar[int]
    PRINCIPAL_TYPE_FIELD_NUMBER: _ClassVar[int]
    principal_id: str
    principal_type: _principal__client_pb2.PrincipalType
    def __init__(self, principal_type: _Optional[_Union[_principal__client_pb2.PrincipalType, str]] = ..., principal_id: _Optional[str] = ...) -> None: ...

class RemovePrincipalGroupMembersRequest(_message.Message):
    __slots__ = ["id", "members"]
    ID_FIELD_NUMBER: _ClassVar[int]
    MEMBERS_FIELD_NUMBER: _ClassVar[int]
    id: str
    members: _containers.RepeatedCompositeFieldContainer[RemovePrincipalGroupMember]
    def __init__(self, id: _Optional[str] = ..., members: _Optional[_Iterable[_Union[RemovePrincipalGroupMember, _Mapping]]] = ...) -> None: ...

class RemovePrincipalGroupMembersResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class UpdatePrincipalGroupRequest(_message.Message):
    __slots__ = ["description", "id", "idp_mapping_names", "name"]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    IDP_MAPPING_NAMES_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    description: str
    id: str
    idp_mapping_names: _containers.RepeatedScalarFieldContainer[str]
    name: str
    def __init__(self, id: _Optional[str] = ..., name: _Optional[str] = ..., description: _Optional[str] = ..., idp_mapping_names: _Optional[_Iterable[str]] = ...) -> None: ...

class UpdatePrincipalGroupResponse(_message.Message):
    __slots__ = ["principal_group"]
    PRINCIPAL_GROUP_FIELD_NUMBER: _ClassVar[int]
    principal_group: _principal_group__client_pb2.PrincipalGroup
    def __init__(self, principal_group: _Optional[_Union[_principal_group__client_pb2.PrincipalGroup, _Mapping]] = ...) -> None: ...



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor
INTERNAL_SPARK_CLUSTER_STATUS_CREATING_CLUSTER: InternalSparkClusterStatusEnum
INTERNAL_SPARK_CLUSTER_STATUS_HEALTHY: InternalSparkClusterStatusEnum
INTERNAL_SPARK_CLUSTER_STATUS_NO_CLUSTER: InternalSparkClusterStatusEnum
INTERNAL_SPARK_CLUSTER_STATUS_UNHEALTHY: InternalSparkClusterStatusEnum
INTERNAL_SPARK_CLUSTER_STATUS_UNSPECIFIED: InternalSparkClusterStatusEnum
INTERNAL_SPARK_CLUSTER_STATUS_WAITING_FOR_CLUSTER_TO_START: InternalSparkClusterStatusEnum

class InternalSparkClusterStatus(_message.Message):
    __slots__ = ["cluster_url", "error", "error_message", "status"]
    CLUSTER_URL_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    cluster_url: str
    error: bool
    error_message: str
    status: InternalSparkClusterStatusEnum
    def __init__(self, status: _Optional[_Union[InternalSparkClusterStatusEnum, str]] = ..., error: bool = ..., error_message: _Optional[str] = ..., cluster_url: _Optional[str] = ...) -> None: ...

class InternalSparkClusterStatusEnum(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class StatusEntry(_message.Message):
    __slots__ = ["anchor_time", "raw_data_end_time", "source_type"]
    ANCHOR_TIME_FIELD_NUMBER: _ClassVar[int]
    RAW_DATA_END_TIME_FIELD_NUMBER: _ClassVar[int]
    SOURCE_TYPE_FIELD_NUMBER: _ClassVar[int]
    anchor_time: int
    raw_data_end_time: int
    source_type: str
    def __init__(self, source_type: _Optional[str] = ..., raw_data_end_time: _Optional[int] = ..., anchor_time: _Optional[int] = ...) -> None: ...



from tecton_proto.args import basic_info__client_pb2 as _basic_info__client_pb2
from tecton_proto.args import diff_options__client_pb2 as _diff_options__client_pb2
from tecton_proto.args import user_defined_function__client_pb2 as _user_defined_function__client_pb2
from tecton_proto.common import id__client_pb2 as _id__client_pb2
from tecton_proto.common import secret__client_pb2 as _secret__client_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ResourceProviderArgs(_message.Message):
    __slots__ = ["function", "info", "prevent_destroy", "resource_provider_id", "secrets"]
    class SecretsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _secret__client_pb2.SecretReference
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_secret__client_pb2.SecretReference, _Mapping]] = ...) -> None: ...
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    INFO_FIELD_NUMBER: _ClassVar[int]
    PREVENT_DESTROY_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_PROVIDER_ID_FIELD_NUMBER: _ClassVar[int]
    SECRETS_FIELD_NUMBER: _ClassVar[int]
    function: _user_defined_function__client_pb2.UserDefinedFunction
    info: _basic_info__client_pb2.BasicInfo
    prevent_destroy: bool
    resource_provider_id: _id__client_pb2.Id
    secrets: _containers.MessageMap[str, _secret__client_pb2.SecretReference]
    def __init__(self, resource_provider_id: _Optional[_Union[_id__client_pb2.Id, _Mapping]] = ..., info: _Optional[_Union[_basic_info__client_pb2.BasicInfo, _Mapping]] = ..., secrets: _Optional[_Mapping[str, _secret__client_pb2.SecretReference]] = ..., function: _Optional[_Union[_user_defined_function__client_pb2.UserDefinedFunction, _Mapping]] = ..., prevent_destroy: bool = ...) -> None: ...
